"""Utility functions for PyTorch models."""

from .stochastic import GaussianSample
from .distributions import (
    log_normal_standard,
    log_normal_diag,
    log_gaussian
)
from .utils import (
    to_categorical,
    get_empty_traces,
    get_empty_dicts,
    LogConfusionMatrix
)
from .metrics import *
from .losses import *

__all__ = [
    # Stochastic
    "GaussianSample",
    
    # Distributions
    "log_normal_standard",
    "log_normal_diag",
    "log_gaussian",
    
    # Utils
    "to_categorical",
    "get_empty_traces",
    "get_empty_dicts",
    "LogConfusionMatrix"
]



import math
import torch
import torch.nn.functional as F


def log_standard_categorical(p):
    """
    Calculates the cross entropy between a (one-hot) categorical vector
    and a standard (uniform) categorical distribution.

    :param p: one-hot categorical distribution
    :return: H(p, u)
    """
    # Uniform prior over y
    prior = F.softmax(torch.ones_like(p), dim=1)
    prior.requires_grad = False

    cross_entropy = -torch.sum(p * torch.log(prior + 1e-8), dim=1)

    return cross_entropy


def log_bernoulli(x, p, eps=0.0):
    """
    Compute log pdf of a Bernoulli distribution with success probability p, at values x.
        .. math:: \log p(x; p) = \log \mathcal{B}(x; p)
    Parameters
    ----------
    x : torch tensor
        Values at which to evaluate pdf.
    p : torch tensor
        Success probability :math:`p(x=1)`, which is also the mean of the Bernoulli distribution.
    eps : float
        Small number used to avoid NaNs by clipping p in range [eps;1-eps].
    Returns
    -------
    torch tensor
        Element-wise log probability, this has to be summed for multi-variate distributions.
    """
    p = torch.clip(p, eps, 1.0 - eps)
    return -F.binary_cross_entropy(p, x)


def log_standard_gaussian(x):
    """
    Evaluates the log pdf of a standard normal distribution at x.

    :param x: point to evaluate
    :return: log N(x|0,I)
    """
    return torch.sum(-0.5 * math.log(2 * math.pi) - x ** 2 / 2, dim=-1)


def log_gaussian(x, mu, log_var):
    """
    Returns the log pdf of a normal distribution parametrised
    by mu and log_var evaluated at x.

    :param x: point to evaluate
    :param mu: mean of distribution
    :param log_var: log variance of distribution
    :return: log N(x|µ,σ)
    """
    log_pdf = - 0.5 * torch.log(2 * torch.tensor(math.pi, requires_grad=True)) - log_var / 2 - (x - mu) ** 2 / (
        2 * torch.exp(log_var))
    return torch.sum(log_pdf, dim=-1)


def log_normal_diag(x, mean, log_var, average=False, dim=1):
    log_normal = -0.5 * (log_var + torch.pow(x - mean, 2) / torch.exp(log_var))
    if average:
        return torch.mean(log_normal, dim)
    else:
        return torch.sum(log_normal, dim)


def log_normal_standard(x, average=False, dim=1):
    log_normal = -0.5 * torch.pow(x, 2)
    if average:
        return torch.mean(log_normal, dim)
    else:
        return torch.sum(log_normal, dim)





# os.environ['CUDA_VISIBLE_DEVICES'] = '-1'
# CUDA_VISIBLE_DEVICES = ""
import os

from tensorboard.plugins.hparams import api as hp
import tensorflow as tf
from sklearn.metrics import silhouette_score, adjusted_rand_score, adjusted_mutual_info_score
from bernn.dl.models.pytorch.utils.metrics import rKBET, rLISI
import numpy as np
import mlflow
import pandas as pd
import matplotlib
import shap

from bernn.utils.utils import get_unique_labels

matplotlib.use('Agg')
import torch
import matplotlib.pyplot as plt
from matplotlib import rcParams, cycler
from matplotlib.lines import Line2D
from .plotting import confidence_ellipse
from sklearn.preprocessing import OneHotEncoder
from sklearn.neighbors import KNeighborsClassifier
from .metrics import batch_f1_score
from .utils import save_roc_curve, save_precision_recall_curve
import seaborn as sns
from sklearn.decomposition import PCA
from sklearn.cross_decomposition import CCA

# try:
#     from umap import UMAP
# except:
from umap.umap_ import UMAP

# from sklearn.manifold import TSNE
# from bernn.utils.metrics import calculate_aic, calculate_bic
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as LDA

# It is useless to run tensorflow on GPU and it takes a lot of GPU RAM for nothing
physical_devices = tf.config.list_physical_devices('CPU')
tf.config.set_visible_devices(physical_devices)


class TensorboardLoggingAE:
    def __init__(self, hparams_filepath, params, variational, zinb, tw, tl, dloss, pseudo,
                 train_after_warmup, berm, args):
        self.params = params
        self.train_after_warmup = train_after_warmup
        self.tw = tw
        self.berm = berm
        self.tl = tl
        self.dloss = dloss
        self.pseudo = pseudo
        self.variational = variational
        self.zinb = zinb
        # self.dann_sets = dann_sets
        # self.dann_plates = dann_batches
        self.hparams_filepath = hparams_filepath
        # self.mz_bin = args.mz_bin
        # self.rt_bin = args.rt_bin
        # self.mzp_bin = args.mz_bin_post
        # self.rtp_bin = args.rt_bin_post
        # self.spd = args.spd
        # self.ms = args.ms_level
        # self.shift = args.shift
        self.berm = args.berm
        # self.log = args.log2
        # self.featselect = args.feature_selection
        self.rec_loss = args.rec_loss
        self.dataset_name = args.dataset
        self.model_name = args.model_name
        self.groupkfold = args.groupkfold
        self.strategy = args.strategy
        self.bad_batches = args.bad_batches
        self.remove_zeros = args.remove_zeros
        self.use_mapping = args.use_mapping
        self.n_emb = args.embeddings_meta
        self.n_meta = args.n_meta
        self.csv_file = args.csv_file
        self.tied_weights = args.tied_weights
        HPARAMS = [
            # hp.HParam('mz_bin', hp.RealInterval(0.0, 100.0)),
            # hp.HParam('rt_bin', hp.RealInterval(0.0, 100.0)),
            # hp.HParam('mzp_bin', hp.RealInterval(0.0, 10000.0)),
            # hp.HParam('rtp_bin', hp.RealInterval(0.0, 10000.0)),
            # hp.HParam('spd', hp.IntInterval(0, 1000)),
            # hp.HParam('ms', hp.Discrete([0, 1])),
            # hp.HParam('shift', hp.Discrete([0, 1])),
            hp.HParam('berm', hp.Discrete(['none', 'combat', 'harmony'])),
            # hp.HParam('log', hp.Discrete(['inloop', 'after', 'none'])),
            hp.HParam('rec', hp.Discrete(['mse', 'l1', 'bce'])),
            # hp.HParam('featselect', hp.Discrete(['mutual_info_classif', 'f_classif'])),

            # hp.HParam('thres', hp.RealInterval(0.0, 1.0)),
            hp.HParam('dropout', hp.RealInterval(0.0, 1.0)),
            hp.HParam('smoothing', hp.RealInterval(0.0, 1.0)),
            hp.HParam('margin', hp.RealInterval(0.0, 10.0)),
            hp.HParam('gamma', hp.RealInterval(0., 100.0)),
            hp.HParam('beta', hp.RealInterval(0., 100.0)),
            hp.HParam('zeta', hp.RealInterval(0., 100.0)),
            hp.HParam('nu', hp.RealInterval(0., 100.0)),
            hp.HParam('layer1', hp.IntInterval(0, 256)),
            hp.HParam('layer2', hp.IntInterval(0, 1024)),
            # hp.HParam('ncols', hp.IntInterval(0, 10000)),
            hp.HParam('lr', hp.RealInterval(1e-8, 1e-2)),
            hp.HParam('wd', hp.RealInterval(1e-15, 1e-1)),
            hp.HParam('scale', hp.Discrete(['binarize', 'minmax', 'standard', 'robust'])),
            hp.HParam('dloss', hp.Discrete(['normae', 'DANN', 'revTriplet', 'inverseTriplet'])),
            # hp.HParam('dann_sets', hp.Discrete([0, 1])),
            # hp.HParam('dann_plates', hp.Discrete([0, 1])),
            hp.HParam('zinb', hp.Discrete([0, 1])),
            hp.HParam('variational', hp.Discrete([0, 1])),
            hp.HParam('tied_w', hp.Discrete([0, 1])),
            hp.HParam('pseudo', hp.Discrete([0, 1])),
            hp.HParam('tripletloss', hp.Discrete([0, 1])),
            hp.HParam('tripletdloss', hp.Discrete([0, 1])),
            hp.HParam('train_after_warmup', hp.Discrete([0, 1]))
        ]

        metrics = [
            hp.Metric('rec_loss', display_name='Rec Loss'),
            hp.Metric('dom_loss', display_name='Domain Loss'),
            hp.Metric('dom_acc', display_name='Domain Accuracy'),
            hp.Metric('train/loss', display_name='Train Loss'),
            hp.Metric('valid/loss', display_name='Valid Loss'),
            hp.Metric('test/loss', display_name='Test Loss'),
            hp.Metric('train/acc', display_name='Train Accuracy'),
            hp.Metric('valid/acc', display_name='Valid Accuracy'),
            hp.Metric('test/acc', display_name='Test Accuracy'),
            hp.Metric('train/top3', display_name='Train top3'),
            hp.Metric('valid/top3', display_name='Valid top3'),
            hp.Metric('test/top3', display_name='Test top3'),
            hp.Metric('train/mcc', display_name='Train MCC'),
            hp.Metric('valid/mcc', display_name='Valid MCC'),
            hp.Metric('test/mcc', display_name='Test MCC'),
            hp.Metric('enc b_euclidean/tot_eucl', display_name='enc b_euclidean/tot_eucl'),
            hp.Metric('enc qc_dist/tot_eucl', display_name='enc qc_dist/tot_eucl'),
            hp.Metric('enc qc_aPCC', display_name='enc qc_aPCC'),
            hp.Metric('enc batch_entropy', display_name='enc qc_aPCC'),
            hp.Metric('rec b_euclidean/tot_eucl', display_name='rec b_euclidean/tot_eucl'),
            hp.Metric('rec qc_dist/tot_eucl', display_name='rec qc_dist/tot_eucl'),
            hp.Metric('rec qc_aPCC', display_name='rec qc_aPCC'),
            hp.Metric('rec batch_entropy', display_name='enc qc_aPCC'),
        ]

        self.writer = tf.summary.create_file_writer(hparams_filepath)
        with self.writer.as_default():
            hp.hparams_config(
                hparams=HPARAMS,
                metrics=metrics,
            )
        self.writer.flush()

    def logging(self, traces, metrics):
        with self.writer.as_default():
            hp.hparams({
                'berm': self.berm,
                'rec': self.rec_loss,
                'gamma': self.params['gamma'],
                'beta': self.params['beta'],
                'zeta': self.params['zeta'],
                'nu': self.params['nu'],
                'dropout': self.params['dropout'],
                'smoothing': self.params['smoothing'],
                'lr': self.params['lr'],
                'margin': self.params['margin'],
                'wd': self.params['wd'],
                'layer1': self.params['layer1'],
                'layer2': self.params['layer2'],
                'scale': self.params['scaler'],
                'tied_w': self.tw,
                'tripletloss': self.tl,
                'dloss': self.dloss,
                'variational': self.variational,
                'zinb': self.zinb,
                'pseudo': self.pseudo,
                'train_after_warmup': self.train_after_warmup,
                'rec_loss': self.rec_loss,
                'dataset_name': self.dataset_name,
                'model_name': self.model_name,
                'groupkfold': self.groupkfold,
                'strategy': self.strategy,
                'bad_batches': self.bad_batches,
                'remove_zeros': self.remove_zeros,
                'use_mapping': self.use_mapping,
                'n_emb': self.n_emb,
                'n_meta': self.n_meta,
                'csv_file': self.csv_file,
                'tied_weights': self.tied_weights
            })  # record the values used in this trial
            tf.summary.scalar('rec_loss', traces['rec_loss'], step=1)
            tf.summary.scalar('dom_loss', traces['dom_loss'], step=1)
            try:
                tf.summary.scalar('dom_acc', traces['dom_acc'], step=1)
            except:
                tf.summary.scalar('dom_acc', traces['dom_acc'][0], step=1)

            # tf.summary.scalar('qc_aPCC', metrics['pool_metrics']['encoded']['all']['qc_aPCC'], step=1)
            # tf.summary.scalar('qc_aPCC_input', metrics['pool_metrics']['inputs']['all']['qc_aPCC'], step=1)
            # tf.summary.scalar('qc_aPCC_rec', metrics['pool_metrics']['recs']['all']['qc_aPCC'], step=1)
            # tf.summary.scalar('qc_dist', metrics['pool_metrics']['encoded']['all']['qc_dist'], step=1)
            # tf.summary.scalar('qc_dist_input', metrics['pool_metrics']['inputs']['all']['qc_dist'], step=1)
            # tf.summary.scalar('qc_dist_rec', metrics['pool_metrics']['recs']['all']['qc_dist'], step=1)

            for g in ['train', 'valid', 'test']:
                tf.summary.scalar(f'{g}/loss', traces[f'{g}_loss'], step=1)
                tf.summary.scalar(f'{g}/acc', traces[f'{g}_acc'], step=1)
                tf.summary.scalar(f'{g}/top3', traces[f'{g}_top3'], step=1)
                tf.summary.scalar(f'{g}/mcc', traces[f'{g}_mcc'], step=1)
            tf.summary.scalar('enc batch_entropy', metrics['batches']['all']['batch_entropy']['enc']['domains'][0], step=1),
            tf.summary.scalar('enc b_euclidean/tot_eucl', traces['enc b_euclidean/tot_eucl'], step=1),
            tf.summary.scalar('enc qc_dist/tot_eucl', traces['enc qc_dist/tot_eucl'], step=1),
            tf.summary.scalar('enc qc_aPCC', traces['enc qc_aPCC'], step=1),
            tf.summary.scalar('rec batch_entropy', metrics['batches']['all']['batch_entropy']['rec']['domains'][0], step=1),
            tf.summary.scalar('rec b_euclidean/tot_eucl', traces['rec b_euclidean/tot_eucl'], step=1),
            tf.summary.scalar('rec qc_dist/tot_eucl', traces['rec qc_dist/tot_eucl'], step=1),
            tf.summary.scalar('rec qc_aPCC', traces['rec qc_aPCC'], step=1),
            
            # if 'batches' in metrics:
            #     for metric in list(metrics['batches']['set'].keys()):
            #         if 'F1' in metric:
            #             for repres in ['enc', 'rec', 'inputs']:
            #                 tf.summary.scalar(f'{metric} {repres} set', metrics['batches']['set'][metric][repres]['set'], step=1)
            #                 tf.summary.scalar(f'{metric} {repres} domains', metrics['batches']['set'][metric][repres]['domains'], step=1)
        self.writer.flush()


def batch_entropy(proba):
    prob_list = []
    for prob in proba:
        loc = 0
        for p in prob:
            loc -= p * np.log(p + 1e-8)
        prob_list += [loc]
    return np.mean(prob_list)


def interactions_mean_matrix(shap_interactions, run, group):
    # Get absolute mean of matrices
    mean_shap = np.abs(shap_interactions).mean(0)
    df = pd.DataFrame(mean_shap, index=X.columns, columns=X.columns)

    # times off diagonal by 2
    df.where(df.values == np.diagonal(df), df.values * 2, inplace=True)

    # display
    plt.figure(figsize=(10, 10), facecolor='w', edgecolor='k')
    sns.set(font_scale=1.5)
    sns.heatmap(df, cmap='coolwarm', annot=True, fmt='.3g', cbar=False)
    plt.yticks(rotation=0)
    f = plt.gcf()
    run[f'shap/interactions_matrix/{group}_values'].upload(f)
    plt.close(f)


def make_summary_plot(df, values, group, run, log_path, category='explainer', mlops='mlflow'):
    shap.summary_plot(values, df, show=False)
    f = plt.gcf()
    if mlops == 'neptune':
        run[f'shap/summary_{category}/{group}_values'].upload(f)
    if mlops == 'mlflow':
        os.makedirs(f'{log_path}/shap/summary_{category}', exist_ok=True)
        plt.savefig(f'{log_path}/shap/summary_{category}/{group}_values.png')
        mlflow.log_figure(f, f'{log_path}/shap/summary_{category}/{group}_values.png')

    plt.close(f)


def make_force_plot(df, values, features, group, run, log_path, category='explainer', mlops='mlflow'):
    shap.force_plot(df, values, features=features, show=False)
    f = plt.gcf()
    if mlops == 'neptune':
        run[f'shap/force_{category}/{group}_values'].upload(f)
    if mlops == 'mlflow':
        os.makedirs(f'{log_path}/shap/force_{category}', exist_ok=True)
        plt.savefig(f'{log_path}/shap/force_{category}/{group}_values.png')
        mlflow.log_figure(f, f'{log_path}/shap/force_{category}/{group}_values.png')

    plt.close(f)


def make_deep_beeswarm(df, values, group, run, log_path, category='explainer', mlops='mlflow'):
    shap.summary_plot(values, feature_names=df.columns, features=df, show=False)
    f = plt.gcf()
    if mlops == 'neptune':
        run[f'shap/beeswarm_{category}/{group}_values'].upload(f)
    if mlops == 'mlflow':
        os.makedirs(f'{log_path}/shap/beeswarm_{category}', exist_ok=True)
        plt.savefig(f'{log_path}/shap/beeswarm_{category}/{group}_values.png')
        mlflow.log_figure(f, f'{log_path}/shap/beeswarm_{category}/{group}_values.png')

    plt.close(f)


def make_decision_plot(df, values, misclassified, feature_names, group, run, log_path, category='explainer', mlops='mlflow'):
    shap.decision_plot(df, values, feature_names=list(feature_names), show=False, link='logit', highlight=misclassified)
    f = plt.gcf()
    if mlops == 'neptune':
        run[f'shap/decision_{category}/{group}_values'].upload(f)
        run[f'shap/decision_{category}/{group}_values'].upload(f)
    if mlops == 'mlflow':
        os.makedirs(f'{log_path}/shap/decision_{category}', exist_ok=True)
        plt.savefig(f'{log_path}/shap/decision_{category}/{group}_values.png')
        mlflow.log_figure(f, f'{log_path}/shap/decision_{category}/{group}_values.png')
    plt.close(f)


def make_decision_deep(df, values, misclassified, feature_names, group, run, log_path, category='explainer', mlops='mlflow'):
    shap.decision_plot(df, values, feature_names=list(feature_names), show=False, link='logit', highlight=misclassified)
    f = plt.gcf()
    if mlops == 'neptune':
        run[f'shap/decision_{category}/{group}_values'].upload(f)
        run[f'shap/decision_{category}/{group}_values'].upload(f)
    if mlops == 'mlflow':
        os.makedirs(f'{log_path}/shap/decision_{category}', exist_ok=True)
        plt.savefig(f'{log_path}/shap/decision_{category}/{group}_values.png')
        mlflow.log_figure(f, f'{log_path}/shap/decision_{category}/{group}_values.png')
    plt.close(f)


def make_multioutput_decision_plot(df, values, group, run, log_path, category='explainer', mlops='mlflow'):
    shap.multioutput_decision_plot(values, df, show=False)
    f = plt.gcf()
    if mlops == 'neptune':
        run[f'shap/multioutput_decision_{category}/{group}_values'].upload(f)
    if mlops == 'mlflow':
        os.makedirs(f'{log_path}/shap/multioutput_decision_{category}', exist_ok=True)
        plt.savefig(f'{log_path}/shap/multioutput_decision_{category}/{group}_values.png')
        mlflow.log_figure(f, f'{log_path}/shap/multioutput_decision_{category}/{group}_values.png')
    plt.close(f)


def make_group_difference_plot(values, mask, group, run, log_path, category='explainer', mlops='mlflow'):
    shap.group_difference_plot(values, mask, show=False)
    f = plt.gcf()
    if mlops == 'neptune':
        # run[f'shap/gdiff_{category}/{group}'].upload(f)
        run[f'shap/gdiff_{category}/{group}'].upload(f)
    if mlops == 'mlflow':
        os.makedirs(f'{log_path}/shap/gdiff_{category}', exist_ok=True)
        plt.savefig(f'{log_path}/shap/gdiff_{category}/{group}_values.png')
        mlflow.log_figure(f, f'{log_path}/shap/gdiff_{category}/{group}_values.png')
    plt.close(f)


def make_beeswarm_plot(values, group, run, log_path, category='explainer', mlops='mlflow'):
    shap.plots.beeswarm(values, max_display=20, show=False)
    f = plt.gcf()
    if mlops == 'neptune':
        # run[f'shap/beeswarm_{category}/{group}'].upload(f)
        run[f'shap/beeswarm_{category}/{group}'].upload(f)
    if mlops == 'mlflow':
        os.makedirs(f'{log_path}/shap/beeswarm_{category}', exist_ok=True)
        plt.savefig(f'{log_path}/shap/beeswarm_{category}/{group}_values.png')
        mlflow.log_figure(f, f'{log_path}/shap/beeswarm_{category}/{group}_values.png')
    plt.close(f)


def make_heatmap(values, group, run, log_path, category='explainer', mlops='mlflow'):
    shap.plots.heatmap(values, instance_order=values.values.sum(1).argsort(), max_display=20, show=False)
    f = plt.gcf()
    if mlops == 'neptune':
        # run[f'shap/heatmap_{category}/{group}'].upload(f)
        run[f'shap/heatmap_{category}/{group}'].upload(f)
    if mlops == 'mlflow':
        os.makedirs(f'{log_path}/shap/heatmap_{category}', exist_ok=True)
        plt.savefig(f'{log_path}/shap/heatmap_{category}/{group}_values.png')
        mlflow.log_figure(f, f'{log_path}/shap/heatmap_{category}/{group}_values.png')
    plt.close(f)


def make_heatmap_deep(values, group, run, log_path, category='explainer', mlops='mlflow'):

    shap.plots.heatmap(pd.DataFrame(values), instance_order=values.sum(1).argsort(), max_display=20, show=False)
    f = plt.gcf()
    if mlops == 'neptune':
        # run[f'shap/heatmap_{category}/{group}'].upload(f)
        run[f'shap/heatmap_{category}/{group}'].upload(f)
    if mlops == 'mlflow':
        os.makedirs(f'{log_path}/shap/heatmap_{category}', exist_ok=True)
        plt.savefig(f'{log_path}/shap/heatmap_{category}/{group}_values.png')
        mlflow.log_figure(f, f'{log_path}/shap/heatmap_{category}/{group}_values.png')
    plt.close(f)


def make_barplot(df, y, values, group, run, log_path, category='explainer', mlops='mlflow'):
    clustering = shap.utils.hclust(df, y, metric='correlation')  # cluster_threshold=0.9
    # shap.plots.bar(values, max_display=20, show=False, clustering=clustering)
    shap.plots.bar(values, max_display=20, show=False, clustering=clustering, clustering_cutoff=0.5)
    f = plt.gcf()
    if mlops == 'neptune':
        run[f'shap/bar_{category}/{group}'].upload(f)
    if mlops == 'mlflow':
        os.makedirs(f'{log_path}/shap/bar_{category}', exist_ok=True)
        plt.savefig(f'{log_path}/shap/bar_{category}/{group}_values.png')
        mlflow.log_figure(f, f'{log_path}/shap/bar_{category}/{group}_values.png')
    plt.close(f)


def make_bar_plot(df, values, group, run, log_path, category='explainer', mlops='mlflow'):
    shap.bar_plot(values, max_display=40, feature_names=df.columns, show=False)
    f = plt.gcf()
    if mlops == 'neptune':
        # run[f'shap/barold_{category}/{group}'].upload(f)
        run[f'shap/barold_{category}/{group}'].upload(f)
    if mlops == 'mlflow':
        os.makedirs(f'{log_path}/shap/barold_{category}', exist_ok=True)
        plt.savefig(f'{log_path}/shap/barold_{category}/{group}_values.png')
        mlflow.log_figure(f, f'{log_path}/shap/barold_{category}/{group}_values.png')
    plt.close(f)


def make_dependence_plot(df, values, var, group, run, log_path, category='explainer', mlops='mlflow'):
    shap.dependence_plot(var, values[1], df, show=False)
    f = plt.gcf()
    if mlops == 'neptune':
        run[f'shap/dependence_{category}/{group}'].upload(f)
    if mlops == 'mlflow':
        os.makedirs(f'{log_path}/shap/dependence_{category}', exist_ok=True)
        plt.savefig(f'{log_path}/shap/dependence_{category}/{group}_values.png')
        mlflow.log_figure(f, f'{log_path}/shap/dependence_{category}/{group}_values.png')
    plt.close(f)


def log_explainer(model, x_df, labels, group, run, cats, log_path, device):
    unique_labels = np.unique(labels)
    # The explainer doesn't like tensors, hence the f function
    explainer = shap.Explainer(model.to(device), x_df, max_evals=2 * x_df.shape[1] + 1)

    # Get the shap values from my test data
    shap_values = explainer(x_df)

    for i, label in enumerate(unique_labels):
        if i == len(shap_values):
            break
        shap_values_df = pd.DataFrame(shap_values[i].data.reshape(1, -1), columns=x_df.columns)
        try:
            shap_values_df.to_csv(f"{log_path}/{group}_permutation_shap_{label}.csv")
        except:
            pass

    make_barplot(x_df, labels, shap_values[:, :, 0], group, run, 'Kernel')

    # Summary plot
    make_summary_plot(x_df, shap_values[:, :, 0], group, run, 'PermutExplainer')

    make_beeswarm_plot(shap_values[:, :, 0], group, run, 'PermutExplainer')
    make_heatmap(shap_values[:, :, 0], group, run, 'PermutExplainer')

    mask = np.array([np.argwhere(x[0] == 1)[0][0] for x in cats])
    make_group_difference_plot(x_df.sum(1).to_numpy(), mask, group, run, 'PermutExplainer')


def log_deep_explainer(model, x_df, misclassified, labels, group, run, cats, log_path, mlops, device):
    unique_labels = np.unique(labels)
    # The explainer doesn't like tensors, hence the f function
    explainer = shap.DeepExplainer(model.to(device), torch.Tensor(x_df.values).to(device))

    # Get the shap values from my test data
    shap_values = explainer.shap_values(torch.Tensor(x_df.values).to(device))

    # Summary plot
    make_summary_plot(x_df, shap_values, group, run, log_path, 'DeepExplainer', mlops)
    make_force_plot(explainer.expected_value[0], shap_values[0][0], x_df.columns, group, run, log_path, 'DeepExplainer', mlops)
    make_deep_beeswarm(x_df, shap_values[0], group, run, log_path, 'DeepExplainer', mlops)
    make_decision_deep(explainer.expected_value[0], shap_values[0], misclassified, x_df.columns, group, run, 'DeepExplainer')

    for i, label in enumerate(unique_labels):
        if i == len(shap_values):
            break
        shap_values_df = pd.DataFrame(shap_values[i], columns=x_df.columns, index=x_df.index)
        try:
            shap_values_df.to_csv(f"{log_path}/{group}_deep_shap_{label}.csv")
        except:
            pass

    try:
        make_dependence_plot(x_df, shap_values, 'APOE', group, run, log_path, 'DeepExplainer', mlops)
    except:
        pass

    # mask = np.array([np.argwhere(x[0] == 1)[0][0] for x in cats])
    # make_group_difference_plot(x_df, mask, group, run, 'DeepExplainer')


def log_kernel_explainer(model, x_df, misclassified, labels, group, run, cats, log_path):
    unique_labels = np.unique(labels)

    f = lambda x: model.to('cpu')(torch.from_numpy(x)).detach().cpu().numpy()

    # Convert my pandas dataframe to numpy
    data = x_df.to_numpy(dtype=np.float32)

    # The explainer doesn't like tensors, hence the f function
    explainer = shap.KernelExplainer(f, data)

    # Get the shap values from my test data
    df = pd.DataFrame(data, columns=x_df.columns)
    shap_values = explainer.shap_values(df)
    # shap_interaction = explainer.shap_interaction_values(X_test)
    shap_values_df = pd.DataFrame(np.concatenate(shap_values), columns=x_df.columns)
    for i, label in enumerate(unique_labels):
        if i == len(shap_values):
            break
        shap_values_df.iloc[i].to_csv(f"{log_path}/{group}_kernel_shap_{label}.csv")
    # shap_values = pd.DataFrame(np.concatenate(s))
    # Summary plot
    make_summary_plot(x_df, shap_values, group, run, 'Kernel')

    make_bar_plot(x_df, shap_values_df.iloc[1], group, run, 'localKernel')

    make_decision_plot(explainer.expected_value[0], shap_values[0], misclassified, x_df.columns, group, run, 'Kernel')

    mask = np.array([np.argwhere(x[0] == 1)[0][0] for x in cats])
    make_group_difference_plot(x_df.sum(1).to_numpy(), mask, group, run, 'Kernel')


def log_shap(run, ae, best_lists, cols, n_meta, mlops, log_path, device, log_deep_only=True):
    # explain all the predictions in the test set
    # explainer = shap.KernelExplainer(svc_linear.predict_proba, X_train[:100])
    os.makedirs(log_path, exist_ok=True)
    for group in ['valid', 'test']:
        if n_meta > 0:
            X = np.concatenate((
                np.concatenate(best_lists[group]['inputs']),
                np.concatenate(best_lists[group]['age']).reshape(-1, 1),
                np.concatenate(best_lists[group]['gender']).reshape(-1, 1),
            ), 1)
            X_test = torch.Tensor(X).to(device)
            X_test_df = pd.DataFrame(X, columns=list(cols) + ['age', 'sex'])
        else:
            X = np.concatenate(best_lists[group]['inputs'])
            X_test = torch.Tensor(X).to(device)
            X_test_df = pd.DataFrame(X, columns=list(cols))

        # explainer = shap.DeepExplainer(ae, X_test)
        # explanation = shap.Explanation(X_test, feature_names=X_test_df.columns)
        # explanation.values = explanation.values.detach().cpu().numpy()
        misclassified = [pred != label for pred, label in zip(np.concatenate(best_lists[group]['preds']).argmax(1),
                                                              np.concatenate(best_lists[group]['cats']).argmax(1))]
        log_deep_explainer(ae, X_test_df, misclassified, np.concatenate(best_lists[group]['labels']),
                           group, run, best_lists[group]['cats'], log_path, mlops, device
                           )
        if not log_deep_only:
            # TODO Problem with not enough memory...
            log_explainer(ae, X_test_df, np.concatenate(best_lists[group]['labels']),
                          group, run, best_lists[group]['cats'], log_path, device
                          )
            log_kernel_explainer(ae, X_test_df,
                                 misclassified,
                                 np.concatenate(best_lists[group]['labels']),
                                 group, run, best_lists[group]['cats'], log_path
                                 )


def log_neptune(run, traces):
    if not np.isnan(traces['rec_loss']):
        run["rec_loss"].log(traces['rec_loss'])
    if not np.isnan(traces['dom_loss']):
        run["dom_loss"].log(traces['dom_loss'])
    if not np.isnan(traces['dom_acc']):
        try:
            run["dom_acc"].log(traces['dom_acc'])
        except:
            run["dom_acc"].log(traces['dom_acc'][0])

    # tf.summary.scalar('qc_aPCC', metrics['pool_metrics']['encoded']['all']['qc_aPCC'], step=1)
    # tf.summary.scalar('qc_aPCC_input', metrics['pool_metrics']['inputs']['all']['qc_aPCC'], step=1)
    # tf.summary.scalar('qc_aPCC_rec', metrics['pool_metrics']['recs']['all']['qc_aPCC'], step=1)
    # tf.summary.scalar('qc_dist', metrics['pool_metrics']['encoded']['all']['qc_dist'], step=1)
    # tf.summary.scalar('qc_dist_input', metrics['pool_metrics']['inputs']['all']['qc_dist'], step=1)
    # tf.summary.scalar('qc_dist_rec', metrics['pool_metrics']['recs']['all']['qc_dist'], step=1)

    for g in ['train', 'valid', 'test']:
        run[f'{g}/loss'].log(traces[f'{g}_loss'])
        run[f'{g}/acc'].log(traces[f'{g}_acc'])
        run[f'{g}/top3'].log(traces[f'{g}_top3'])
        run[f'{g}/mcc'].log(traces[f'{g}_mcc'])
    try:
        run['enc b_euclidean/tot_eucl'].log(traces['enc b_euclidean/tot_eucl']),
        run['enc qc_dist/tot_eucl'].log(traces['enc qc_dist/tot_eucl']),
        run['enc qc_aPCC'].log(traces['enc qc_aPCC']),
        run['enc batch_entropy'].log(traces['enc batch_entropy']),
        run['rec b_euclidean/tot_eucl'].log(traces['rec b_euclidean/tot_eucl']),
        run['rec qc_dist/tot_eucl'].log(traces['rec qc_dist/tot_eucl']),
        run['rec qc_aPCC'].log(traces['rec qc_aPCC']),
        run['rec batch_entropy'].log(traces['rec batch_entropy']),
    except:
        pass


def log_mlflow(traces, step):
    if 'rec_loss' in traces:
        if not np.isnan(traces['rec_loss']):
            try:
                mlflow.log_metric("rec_loss", traces['rec_loss'], step)
            except:
                print(f"\n\n\nPROBLEM HERE:::: {traces['rec_loss']}\n\n\n")
        if not np.isnan(traces['dom_loss']):
            mlflow.log_metric("dom_loss", traces['dom_loss'], step)
        if not np.isnan(traces['dom_acc']):
            try:
                mlflow.log_metric("dom_acc", traces['dom_acc'], step)
            except:
                mlflow.log_metric("dom_acc", traces['dom_acc'], step)

    # tf.summary.scalar('qc_aPCC', metrics['pool_metrics']['encoded']['all']['qc_aPCC'], step=1)
    # tf.summary.scalar('qc_aPCC_input', metrics['pool_metrics']['inputs']['all']['qc_aPCC'], step=1)
    # tf.summary.scalar('qc_aPCC_rec', metrics['pool_metrics']['recs']['all']['qc_aPCC'], step=1)
    # tf.summary.scalar('qc_dist', metrics['pool_metrics']['encoded']['all']['qc_dist'], step=1)
    # tf.summary.scalar('qc_dist_input', metrics['pool_metrics']['inputs']['all']['qc_dist'], step=1)
    # tf.summary.scalar('qc_dist_rec', metrics['pool_metrics']['recs']['all']['qc_dist'], step=1)

    for g in ['train', 'valid', 'test']:
        mlflow.log_metric(f'{g}/loss', traces[f'{g}_loss'], step)
        mlflow.log_metric(f'{g}/acc', traces[f'{g}_acc'], step)
        mlflow.log_metric(f'{g}/top3', traces[f'{g}_top3'], step)
        mlflow.log_metric(f'{g}/mcc', traces[f'{g}_mcc'], step)
    for rep in ['enc', 'rec']:
        for g in ['all', 'train', 'valid', 'test']:
            try:
                mlflow.log_metric(f"{rep} {g} b_euclidean/tot_eucl", traces['pool_metrics'][rep][g]['[b_euclidean/tot_eucl]'], step)
                mlflow.log_metric(f"{rep} {g} qc_dist/tot_eucl", traces['pool_metrics'][rep][g]['[b_euclidean/tot_eucl]'], step)
                mlflow.log_metric(f"{rep} {g} qc_aPCC", traces['pool_metrics'][rep][g]['[b_euclidean/tot_eucl]'], step)
                mlflow.log_metric(f"{rep} {g} batch_entropy", traces['batches'][g]['batch_entropy'][rep]['domains'][-1], step)
                mlflow.log_metric(f"{rep} {g} silhouette", traces['batches'][g]['silhouette'][rep]['domains'][-1], step)
                mlflow.log_metric(f"{rep} {g} lisi", traces['batches'][g]['lisi'][rep]['domains'][-1], step)
                mlflow.log_metric(f"{rep} {g} kbet", traces['batches'][g]['kbet'][rep]['domains'][-1], step)
                mlflow.log_metric(f"{rep} {g} adjusted_rand_score", traces['batches'][g]['adjusted_rand_score'][rep]['domains'][-1], step)
                mlflow.log_metric(f"{rep} {g} adjusted_mutual_info_score", traces['batches'][g]['adjusted_mutual_info_score'][rep]['domains'][-1], step)
            except:
                pass


def log_PAD(lists, values, model):
    pass


def get_metrics(lists, values, model):
    # sets are grouped togheter for a single metric
    # from sklearn.metrics import silhouette_score, adjusted_rand_score, adjusted_mutual_info_score
    # from bernn.dl.models.pytorch.utils.metrics import rLISI
    # from sklearn.neighbors import KNeighborsClassifier
    
    knns = {info: {repres: KNeighborsClassifier(n_neighbors=20) for repres in ['set', 'domains', 'labels', 'times']} for
            info in ['enc', 'rec', 'inputs']}
    # classes = []
    # domains = []
    # sets = []
    # encoded_values = []
    # rec_values = []
    # inputs = []

    num_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    for group in ['all', 'train', 'valid', 'test']:
        if len(lists[group]['inputs']) > 0:
            try:
                classes = np.array(np.concatenate(lists[group]['classes']), np.int)
            except:
                pass

            if group == 'all':
                knns['enc']['domains'].fit(
                    np.concatenate(lists[group]['encoded_values']),
                    np.concatenate(lists[group]['domains'])
                )
                if len(lists[group]['rec_values']) > 0:
                    knns['rec']['domains'].fit(
                        np.concatenate(lists[group]['rec_values']),
                        np.concatenate(lists[group]['domains'])
                    )
                knns['inputs']['domains'].fit(
                    np.concatenate(lists[group]['inputs']),
                    np.concatenate(lists[group]['domains'])
                )

                try:
                    knns['enc']['labels'].fit(
                        np.concatenate(lists[group]['encoded_values']),
                        np.concatenate(lists[group]['classes'])
                    )
                    if len(lists[group]['rec_values']) > 0:
                        knns['rec']['labels'].fit(
                            np.concatenate(lists[group]['rec_values']),
                            np.concatenate(lists[group]['classes'])
                        )
                    knns['inputs']['labels'].fit(
                        np.concatenate(lists[group]['inputs']),
                        np.concatenate(lists[group]['classes'])
                    )
                except:
                    pass

            lists[group]['domain_proba'] = knns['inputs']['domains'].predict_proba(
                np.concatenate(lists[group]['inputs']),
            )
            lists[group]['domain_preds'] = knns['inputs']['domains'].predict(
                np.concatenate(lists[group]['inputs']),
            )
            values[group]['batch_entropy']['inputs']['domains'] = [batch_entropy(lists[group]['domain_proba'])]

            lists[group]['domain_proba'] = knns['enc']['domains'].predict_proba(
                np.concatenate(lists[group]['encoded_values']),
            )
            lists[group]['domain_preds'] = knns['enc']['domains'].predict(
                np.concatenate(lists[group]['encoded_values']),
            )
            values[group]['batch_entropy']['enc']['domains'] = [batch_entropy(lists[group]['domain_proba'])]

            if len(lists[group]['rec_values']) > 0:
                lists[group]['domain_proba'] = knns['rec']['domains'].predict_proba(
                    np.concatenate(lists[group]['rec_values']),
                )
                lists[group]['domain_preds'] = knns['rec']['domains'].predict(
                    np.concatenate(lists[group]['rec_values']),
                )
            values[group]['batch_entropy']['rec']['domains'] = [batch_entropy(lists[group]['domain_proba'])]

            # for metric, funct in zip(['lisi', 'silhouette', 'kbet'], [rLISI, silhouette_score, rKBET]):
            for metric, funct in zip(['lisi', 'silhouette'], [rLISI, silhouette_score]):
                try:
                    values[group][metric]['enc']['labels'] = [funct(np.concatenate(lists[group]['encoded_values']),
                                                                np.concatenate(lists[group]['classes']))]
                except:
                    values[group][metric]['enc']['labels'] = [-1]
                try:
                    values[group][metric]['enc']['domains'] = [funct(np.concatenate(lists[group]['encoded_values']),
                                                                     np.concatenate(lists[group]['domains']))]
                except:
                    values[group][metric]['enc']['domains'] = [-1]

                try:
                    values[group][metric]['rec']['labels'] = [funct(np.concatenate(lists[group]['rec_values']),
                                                                    np.concatenate(lists[group]['classes']))]
                except:
                    values[group][metric]['rec']['labels'] = [-1]

                try:
                    values[group][metric]['rec']['domains'] = [funct(np.concatenate(lists[group]['rec_values']),
                                                                     np.concatenate(lists[group]['domains']))]
                except:
                    values[group][metric]['rec']['domains'] = [-1]

                try:
                    values[group][metric]['inputs']['labels'] = [funct(np.concatenate(lists[group]['inputs']),
                                                                       np.concatenate(lists[group]['classes']))]
                except:
                    values[group][metric]['inputs']['labels'] = [-1]

                try:
                    values[group][metric]['inputs']['domains'] = [funct(np.concatenate(lists[group]['inputs']),
                                                                        np.concatenate(lists[group]['domains']))]
                except:
                    values[group][metric]['inputs']['domains'] = [-1]

            for metric, funct in zip(
                    ['adjusted_rand_score', 'adjusted_mutual_info_score'],
                    [adjusted_rand_score, adjusted_mutual_info_score]):


                try:
                    lists[group]['domain_preds'] = knns['enc']['domains'].predict(
                        np.concatenate(lists[group]['encoded_values']),
                    )
                except:
                    pass

                try:
                    values[group][metric]['enc']['domains'] = [funct(np.concatenate(lists[group]['domains']),
                                                                     lists[group]['domain_preds'])]
                except:
                    pass


                try:
                    values[group][metric]['enc']['labels'] = [funct(np.concatenate(lists[group]['domains']),
                                                                    lists[group]['domain_preds'])]
                except:
                    pass

                # lists[group]['domain_preds'] = knns['enc']['labels'].predict(
                #     np.concatenate(lists[group]['encoded_values']),
                # )

                #################


                try:
                    lists[group]['domain_preds'] = knns['rec']['labels'].predict(
                        np.concatenate(lists[group]['rec_values']),
                    )
                except:
                    pass
                try:
                    values[group][metric]['rec']['domains'] = [funct(np.concatenate(lists[group]['domains']),
                                                                     lists[group]['domain_preds'])]
                except:
                    pass

                try:
                    values[group][metric]['rec']['labels'] = [funct(np.concatenate(lists[group]['labels']),
                                                                    lists[group]['domain_preds'])]
                except:
                    pass

                # lists[group]['domain_preds'] = knns['rec']['labels'].predict(
                #     np.concatenate(lists[group]['rec']),
                # )

                #################

                try:
                    lists[group]['domain_preds'] = knns['inputs']['domains'].predict(
                        np.concatenate(lists[group]['inputs']),
                    )
                except:
                    pass

                try:
                    values[group][metric]['inputs']['domains'] = [funct(np.concatenate(lists[group]['domains']),
                                                                        lists[group]['domain_preds'])]
                except:
                    pass

                try:
                    lists[group]['domain_preds'] = knns['inputs']['labels'].predict(
                        np.concatenate(lists[group]['inputs']),
                    )
                except:
                    pass

                try:
                    values[group][metric]['inputs']['labels'] = [funct(np.concatenate(lists[group]['labels']),
                                                                       lists[group]['domain_preds'])]
                except:
                    pass

    return values


def log_ORD(ordin, logger, data, uniques, mlops, epoch, transductive=False):
    model = ordin['model']
    for f in ['inputs', 'batches', 'labels']:
        for g in ['train', 'valid', 'test']:
            # print(f"{f} {g}")
            try:
                data[f][g] = np.concatenate((data[f][g], data[f][f'{g}_pool']))
            except:
                pass
    if transductive:
        model.fit(np.concatenate((data['inputs']['train'], data['inputs']['valid'], data['inputs']['test'])))
        pcs_train = model.transform(data['inputs']['train'])
        if "transductive" not in ordin['name']:
            ordin['name'] += "_transductive"
    else:
        pcs_train = model.fit_transform(data['inputs']['train'])
    if data['inputs']['valid'] is not None:
        pcs_valid = model.transform(data['inputs']['valid'])
    else:
        pcs_valid = np.array([])
    if data['inputs']['test'] is not None:
        pcs_test = model.transform(data['inputs']['test'])
    else:
        pcs_test = np.array([])
        test_labels = np.array([])

    pcs_train_df = pd.DataFrame(data=pcs_train, columns=['PC 1', 'PC 2'])
    pcs_valid_df = pd.DataFrame(data=pcs_valid, columns=['PC 1', 'PC 2'])
    pcs_test_df = pd.DataFrame(data=pcs_test, columns=['PC 1', 'PC 2'])
    for name in list(uniques.keys()):
        fig = plt.figure(figsize=(12, 12))
        ax = fig.add_subplot(111)

        try:
            ev = model.explained_variance_ratio_
            pc1 = 'Component_1 : ' + str(np.round(ev[0] * 100, decimals=2)) + "%"
            pc2 = 'Component_2 : ' + str(np.round(ev[1] * 100, decimals=2)) + "%"
        except:
            pc1 = 'Component_1'
            pc2 = 'Component_2'

        ax.set_xlabel(pc1, fontsize=15)
        ax.set_ylabel(pc2, fontsize=15)
        ax.set_title(f"2 component {ordin['name']}", fontsize=20)

        num_targets = len(uniques[name])
        cmap = plt.cm.tab20

        cols = cmap(np.linspace(0, 1, len(uniques[name]) + 1))
        colors = rcParams['axes.prop_cycle'] = cycler(color=cols)
        colors_list = {name: [] for name in ['train', 'valid', 'test']}
        data1_list = {name: [] for name in ['train', 'valid', 'test']}
        data2_list = {name: [] for name in ['train', 'valid', 'test']}
        new_labels = {name: [] for name in ['train', 'valid', 'test']}
        new_cats = {name: [] for name in ['train', 'valid', 'test']}

        ellipses = []
        unique_cats_train = np.array([])
        for df_name, df, labels in zip(['train', 'valid', 'test'],
                                       [pcs_train_df, pcs_valid_df, pcs_test_df],
                                       [data[name]['train'], data[name]['valid'], data[name]['test']]):
            for t, target in enumerate(uniques[name]):
                indices_to_keep = [True if x == target else False for x in list(labels)]
                data1 = list(df.loc[indices_to_keep, 'PC 1'])
                new_labels[df_name] += [target for _ in range(len(data1))]
                new_cats[df_name] += [target for _ in range(len(data1))]

                data2 = list(df.loc[indices_to_keep, 'PC 2'])
                data1_list[df_name] += [data1]
                data2_list[df_name] += [data2]
                colors_list[df_name] += [np.array([[cols[t]] * len(data1)])]
                if len(indices_to_keep) > 1 and df_name == 'train_data' or target not in unique_cats_train:
                    unique_cats_train = np.unique(np.concatenate((new_labels[df_name], unique_cats_train)))
                    try:
                        confidence_ellipses = confidence_ellipse(np.array(data1), np.array(data2), ax, 1.5,
                                                                 edgecolor=cols[t],
                                                                 train_set=True)
                        ellipses += [confidence_ellipses[1]]
                    except:
                        pass

        for df_name, marker in zip(list(data1_list.keys()), ['o', 'x', '*']):
            data1_vector = np.hstack([d for d in data1_list[df_name] if len(d) > 0]).reshape(-1, 1)
            colors_vector = np.hstack([d for d in colors_list[df_name] if d.shape[1] > 0]).squeeze()
            data2_vector = np.hstack(data2_list[df_name]).reshape(-1, 1)
            data_colors_vector = np.concatenate((data1_vector, data2_vector, colors_vector), axis=1)
            data2 = data_colors_vector[:, 1]
            col = data_colors_vector[:, 2:]
            data1 = data_colors_vector[:, 0]

            ax.scatter(data1, data2, s=50, alpha=1.0, c=col, label=new_labels[df_name], marker=marker)
            custom_lines = [Line2D([0], [0], color=cmap(x), lw=4) for x in np.linspace(0, 1, len(uniques[name]) + 1)]
            ax.legend(custom_lines, uniques[name].tolist())

        if mlops == "tensorboard":
            logger.add_figure(f'{ordin["name"]}_{name}', fig, epoch)
        elif mlops == "neptune":
            logger[f'{ordin["name"]}/{name}'].upload(fig)
        elif mlops == "mlflow":
            mlflow.log_figure(fig, f'{ordin["name"]}/{name}.png')
        else:
            plt.show()

        plt.close(fig)


def log_LDA(ordin, logger, data, uniques, mlops, epoch):
    for name in list(uniques.keys()):
        fig = plt.figure(figsize=(12, 12))
        ax = fig.add_subplot(111)

        if len(uniques[name]) > 2:
            n_comp = 2
        else:
            n_comp = 1
        model = ordin['model'](n_components=n_comp)

        pcs_train = model.fit_transform(data['inputs']['train'], data[name]['train'])
        pcs_valid = model.transform(data['inputs']['valid'])
        pcs_test = model.transform(data['inputs']['test'])

        if n_comp > 1:
            pcs_train_df = pd.DataFrame(data=pcs_train, columns=['LD1', 'LD2'])
            pcs_valid_df = pd.DataFrame(data=pcs_valid, columns=['LD1', 'LD2'])
            pcs_test_df = pd.DataFrame(data=pcs_test, columns=['LD1', 'LD2'])
        else:
            pcs_train_df = pd.DataFrame(data=pcs_train, columns=['LD1'])
            pcs_valid_df = pd.DataFrame(data=pcs_valid, columns=['LD1'])
            pcs_test_df = pd.DataFrame(data=pcs_test, columns=['LD1'])

        try:
            ev = model.explained_variance_ratio_
            pc1 = 'Component_1 : ' + str(np.round(ev[0] * 100, decimals=2)) + "%"
            pc2 = 'Component_2 : ' + str(np.round(ev[1] * 100, decimals=2)) + "%"
        except:
            pc1 = 'Component_1'
            pc2 = 'Component_2'

        ax.set_xlabel(pc1, fontsize=15)
        ax.set_ylabel(pc2, fontsize=15)
        ax.set_title(f"2 component {ordin['name']}", fontsize=20)

        num_targets = len(uniques[name])
        cmap = plt.cm.tab20

        cols = cmap(np.linspace(0, 1, len(uniques[name]) + 1))
        colors = rcParams['axes.prop_cycle'] = cycler(color=cols)
        colors_list = {name: [] for name in ['train_data', 'valid_data', 'test_data']}
        data1_list = {name: [] for name in ['train_data', 'valid_data', 'test_data']}
        data2_list = {name: [] for name in ['train_data', 'valid_data', 'test_data']}
        new_labels = {name: [] for name in ['train_data', 'valid_data', 'test_data']}
        new_cats = {name: [] for name in ['train_data', 'valid_data', 'test_data']}

        ellipses = []
        unique_cats_train = np.array([])
        for df_name, df, labels in zip(['train_data', 'valid_data', 'test_data'],
                                       [pcs_train_df, pcs_valid_df, pcs_test_df],
                                       [data[name]['train'], data[name]['valid'], data[name]['test']]):
            for t, target in enumerate(uniques[name]):
                indices_to_keep = [True if x == target else False for x in
                                   list(labels)]  # 0 is the name of the column with target values
                data1 = list(df.loc[indices_to_keep, 'LD1'])
                new_labels[df_name] += [target for _ in range(len(data1))]
                new_cats[df_name] += [target for _ in range(len(data1))]

                data1_list[df_name] += [data1]
                colors_list[df_name] += [np.array([[cols[t]] * len(data1)])]
                if n_comp > 1:
                    data2 = list(df.loc[indices_to_keep, 'LD2'])
                    data2_list[df_name] += [data2]
                if len(indices_to_keep) > 1 and df_name == 'train_data' or target not in unique_cats_train:
                    unique_cats_train = np.unique(np.concatenate((new_labels[df_name], unique_cats_train)))
                    if n_comp > 1:
                        try:
                            confidence_ellipses = confidence_ellipse(np.array(data1), np.array(data2), ax, 1.5,
                                                                     edgecolor=cols[t],
                                                                     train_set=True)
                            ellipses += [confidence_ellipses[1]]
                        except:
                            pass

        for df_name, marker in zip(list(data1_list.keys()), ['o', 'x', '*']):
            data1_vector = np.hstack([d for d in data1_list[df_name] if len(d) > 0]).reshape(-1, 1)
            colors_vector = np.hstack([d for d in colors_list[df_name] if d.shape[1] > 0]).squeeze()
            if n_comp > 1:
                data2_vector = np.hstack(data2_list[df_name]).reshape(-1, 1)
                data_colors_vector = np.concatenate((data1_vector, data2_vector, colors_vector), axis=1)
                data1 = data_colors_vector[:, 0]
                data2 = data_colors_vector[:, 1]
                col = data_colors_vector[:, 2:]
                ax.scatter(data1, data2, s=50, alpha=1.0, c=col, label=new_labels[df_name], marker=marker)
            else:
                data_colors_vector = np.concatenate((data1_vector, colors_vector), axis=1)
                data1 = data_colors_vector[:, 0]
                col = data_colors_vector[:, 1:]
                ax.scatter(data1, np.random.random(len(data1)), s=50, alpha=1.0, c=col, label=new_labels[df_name],
                           marker=marker)

            custom_lines = [Line2D([0], [0], color=cmap(x), lw=4) for x in np.linspace(0, 1, len(uniques[name]) + 1)]
            ax.legend(custom_lines, uniques[name].tolist())

        if mlops == "tensorboard":
            logger.add_figure(f'{ordin["name"]}_{name}', fig, epoch)
        elif mlops == "neptune":
            logger[f'{ordin["name"]}/{name}'].upload(fig)
        elif mlops == "mlflow":
            mlflow.log_figure(fig, f'{ordin["name"]}/{name}.png')
        plt.close(fig)
        plt.close()


def log_CCA(ordin, logger, data, uniques, mlops, epoch):
    for name in list(uniques.keys()):
        fig = plt.figure(figsize=(12, 12))
        ax = fig.add_subplot(111)

        model = ordin['model']

        train_cats = OneHotEncoder().fit_transform(
            np.stack([np.argwhere(uniques[name] == x) for x in data[name]['train']]).reshape(-1, 1)
        ).toarray()

        pcs_train, _ = model.fit_transform(data['inputs']['train'], train_cats)
        pcs_valid = model.transform(data['inputs']['valid'])
        pcs_test = model.transform(data['inputs']['test'])

        pcs_train_df = pd.DataFrame(data=pcs_train, columns=['PC 1', 'PC 2'])
        pcs_valid_df = pd.DataFrame(data=pcs_valid, columns=['PC 1', 'PC 2'])
        pcs_test_df = pd.DataFrame(data=pcs_test, columns=['PC 1', 'PC 2'])
        try:
            ev = model.explained_variance_ratio_
            pc1 = 'Component_1 : ' + str(np.round(ev[0] * 100, decimals=2)) + "%"
            pc2 = 'Component_2 : ' + str(np.round(ev[1] * 100, decimals=2)) + "%"
        except:
            pc1 = 'Component_1'
            pc2 = 'Component_2'

        ax.set_xlabel(pc1, fontsize=15)
        ax.set_ylabel(pc2, fontsize=15)
        ax.set_title(f"2 component {ordin['name']}", fontsize=20)

        num_targets = len(uniques[name])
        cmap = plt.cm.tab20

        cols = cmap(np.linspace(0, 1, len(uniques[name]) + 1))
        colors = rcParams['axes.prop_cycle'] = cycler(color=cols)
        colors_list = {name: [] for name in ['train_data', 'valid_data', 'test_data']}
        data1_list = {name: [] for name in ['train_data', 'valid_data', 'test_data']}
        data2_list = {name: [] for name in ['train_data', 'valid_data', 'test_data']}
        new_labels = {name: [] for name in ['train_data', 'valid_data', 'test_data']}
        new_cats = {name: [] for name in ['train_data', 'valid_data', 'test_data']}

        ellipses = []
        unique_cats_train = np.array([])
        for df_name, df, labels in zip(['train_data', 'valid_data', 'test_data'],
                                       [pcs_train_df, pcs_valid_df, pcs_test_df],
                                       [data[name]['train'], data[name]['valid'], data[name]['test']]):
            for t, target in enumerate(uniques[name]):
                indices_to_keep = [True if x == target else False for x in
                                   list(labels)]  # 0 is the name of the column with target values
                data1 = list(df.loc[indices_to_keep, 'PC 1'])
                new_labels[df_name] += [target for _ in range(len(data1))]
                new_cats[df_name] += [target for _ in range(len(data1))]

                data2 = list(df.loc[indices_to_keep, 'PC 2'])
                data1_list[df_name] += [data1]
                data2_list[df_name] += [data2]
                colors_list[df_name] += [np.array([[cols[t]] * len(data1)])]
                if len(indices_to_keep) > 1 and df_name == 'train_data' or target not in unique_cats_train:
                    unique_cats_train = np.unique(np.concatenate((new_labels[df_name], unique_cats_train)))
                    try:
                        confidence_ellipses = confidence_ellipse(np.array(data1), np.array(data2), ax, 1.5,
                                                                 edgecolor=cols[t],
                                                                 train_set=True)
                        ellipses += [confidence_ellipses[1]]
                    except:
                        pass

        for df_name, marker in zip(list(data1_list.keys()), ['o', 'x', '*']):
            data1_vector = np.hstack([d for d in data1_list[df_name] if len(d) > 0]).reshape(-1, 1)
            colors_vector = np.hstack([d for d in colors_list[df_name] if d.shape[1] > 0]).squeeze()
            data2_vector = np.hstack(data2_list[df_name]).reshape(-1, 1)
            data_colors_vector = np.concatenate((data1_vector, data2_vector, colors_vector), axis=1)
            data2 = data_colors_vector[:, 1]
            col = data_colors_vector[:, 2:]
            data1 = data_colors_vector[:, 0]

            ax.scatter(data1, data2, s=50, alpha=1.0, c=col, label=new_labels[df_name], marker=marker)
            custom_lines = [Line2D([0], [0], color=cmap(x), lw=4) for x in np.linspace(0, 1, len(uniques[name]) + 1)]
            ax.legend(custom_lines, uniques[name].tolist())

        if mlops == "tensorboard":
            logger.add_figure(f'{ordin["name"]}_{name}', fig, epoch)
        elif mlops == "neptune":
            logger[f'{ordin["name"]}/{name}'].upload(fig)
        elif mlops == "mlflow":
            mlflow.log_figure(fig, f'{ordin["name"]}/{name}.png')
        plt.close(fig)


def log_metrics(logger, lists, values, model, unique_labels, unique_batches, epoch, mlops, metrics, n_meta_emb=0, device='cuda'):
    # from bernn.dl.models.pytorch.utils.metrics import batch_f1_score
    
    if len(unique_labels) > 2:
        bout = 0
    else:
        bout = 1
    try:
        values = get_metrics(lists, values, model)
    except:
        print("\n\n\nProblem with logging metrics\n\n\n")
    # print('pool metrics logged')
    try:
        for repres in ['enc', 'rec', 'inputs']:
            for metric in ['silhouette', 'lisi']:  # 'kbet', 
                for info in ['labels', 'domains']:
                    for group in ['train', 'valid', 'test']:
                        if metric == 'lisi':
                            if mlops == "tensorboard":
                                try:
                                    logger.add_scalar(f'{metric}/{group}/{repres}/{info}',
                                                      values[group][metric][repres][info][0][0], epoch)
                                except:
                                    logger.add_scalar(f'{metric}/{group}/{repres}/{info}',
                                                      values[group][metric][repres][info][0], epoch)
                            elif mlops == "neptune":
                                try:
                                    logger[f'{metric}/{group}/{repres}/{info}'].log(
                                        values[group][metric][repres][info][0])
                                except:
                                    logger[f'{metric}/{group}/{repres}/{info}'].log(
                                        values[group][metric][repres][info][0][0])
                            elif mlops == "mlflow":
                                try:
                                    mlflow.log_metric(
                                        f'{metric}/{group}/{repres}/{info}',
                                        values[group][metric][repres][info][0], epoch
                                    )
                                except:
                                    mlflow.log_metric(f'{metric}/{group}/{repres}/{info}',
                                        values[group][metric][repres][info][0][0], epoch)

                        elif metric == 'silhouette':
                            if mlops == "tensorboard":
                                try:
                                    logger.add_scalar(f'{metric}/{group}/{repres}/{info}',
                                                      values[group][metric][repres][info][0], epoch)
                                except:
                                    pass
                            elif mlops == "neptune":
                                try:
                                    logger[f'{metric}/{group}/{repres}/{info}'].log(
                                        values[group][metric][repres][info][0])
                                except:
                                    pass
                            elif mlops == "mlflow":
                                try:
                                    mlflow.log_metric(f'{metric}/{group}/{repres}/{info}',
                                        values[group][metric][repres][info][0], epoch)
                                except:
                                    pass

                        elif metric == 'kbet':
                            if mlops == "tensorboard":
                                try:
                                    logger.add_scalar(f'{metric}/{group}/{repres}/{info}',
                                                      values[group][metric][repres][info][0], epoch)
                                except:
                                    pass
                            elif mlops == "neptune":
                                try:
                                    logger[f'{metric}/{group}/{repres}/{info}'].log(
                                        values[group][metric][repres][info][0])
                                except:
                                    pass
                            elif mlops == "mlflow":
                                try:
                                    mlflow.log_metric(f'{metric}/{group}/{repres}/{info}',
                                        values[group][metric][repres][info][0], epoch)
                                except:
                                    pass

                for group in ['train', 'valid', 'test']:
                    if metric == 'lisi':
                        try:
                            values[group][f"F1_{metric}"][repres]['domains'] = batch_f1_score(
                                batch_score=values[group][metric][repres]['domains'][0][0] / len(unique_batches),
                                class_score=values[group][metric][repres]['labels'][0][0] / len(unique_labels),
                            )
                        except:
                            values[group][f"F1_{metric}"][repres]['domains'] = batch_f1_score(
                                batch_score=values[group][metric][repres]['domains'][0] / len(unique_batches),
                                class_score=values[group][metric][repres]['labels'][0] / len(unique_labels),
                            )

                    elif metric == 'silhouette':
                        values[group][f"F1_{metric}"][repres]['domains'] = batch_f1_score(
                            batch_score=(values[group][metric][repres]['domains'][0] + 1) / 2,
                            class_score=(values[group][metric][repres]['labels'][0] + 1) / 2,
                        )
                    elif metric == 'kbet':
                        try:
                            values[group][f"F1_{metric}"][repres]['domains'] = batch_f1_score(
                                batch_score=values[group][metric][repres]['domains'][0] / len(unique_batches),
                                class_score=values[group][metric][repres]['labels'][0] / len(unique_labels),
                            )
                        except:
                            pass
                    if mlops == "tensorboard":
                        logger.add_scalar(f'F1_{metric}/{group}/{repres}/domains',
                                          values[group][f"F1_{metric}"][repres]['domains'],
                                          epoch)
                    elif mlops == "neptune":
                        logger[f'F1_{metric}/{group}/{repres}/domains'].log(
                            values[group][f"F1_{metric}"][repres]['domains'],
                        )
                    elif mlops == "mlflow":
                        mlflow.log_metric(f'F1_{metric}/{group}/{repres}/domains',
                            values[group][f"F1_{metric}"][repres]['domains'], epoch
                        )

    except:
        return metrics

    try:
        for repres in ['enc', 'rec', 'inputs']:
            for metric in ['adjusted_rand_score', 'adjusted_mutual_info_score', 'batch_entropy']:
                for group in ['all', 'train', 'valid', 'test']:  # , 'set'
                    for info in ['labels', 'domains']:
                        if metric == 'batch_entropy' and info == 'labels':
                            continue
                        if mlops == "tensorboard":
                            logger.add_scalar(f'{metric}/{group}/{repres}/{info}',
                                              values[group][metric][repres][info][0],
                                              epoch)
                        elif mlops == "neptune":
                            logger[f'{metric}/{group}/{repres}/{info}'].log(
                                values[group][metric][repres][info][0])
                        elif mlops == "mlflow":
                            mlflow.log_metric(f'{metric}/{group}/{repres}/{info}',
                                values[group][metric][repres][info][0], epoch)

                    if metric != 'batch_entropy':
                        values[group][f"F1_{metric}"][repres]['domains'] = batch_f1_score(
                            batch_score=values[group][metric][repres]['domains'][0],
                            class_score=values[group][metric][repres]['labels'][0],
                        )
                        if mlops == "tensorboard":
                            logger.add_scalar(f'F1_{metric}/{group}/{repres}/domains',
                                              values[group][f"F1_{metric}"][repres]['domains'],
                                              epoch)
                        elif mlops == "neptune":
                            logger[f'F1_{metric}/{group}/{repres}/domains'].log(
                                values[group][f"F1_{metric}"][repres]['domains'])
                        elif mlops == "mlflow":
                            mlflow.log_metric(f'F1_{metric}/{group}/{repres}/domains',
                                values[group][f"F1_{metric}"][repres]['domains'], epoch)
    except:
        print("\n\n\nProblem with Adjusted Rand Index or Adjusted Mutual Info\n\n\n")

    # print('batch mixing metrics logged')
    metrics['batches'] = values
    try:
        clusters = ['labels', 'domains']
        reps = ['enc', 'rec']
        train_lisi_enc = [values['train']['lisi']['enc'][c][0].reshape(-1) for c in clusters]
        train_lisi_rec = [values['train']['lisi']['rec'][c][0].reshape(-1) for c in clusters]
        valid_lisi_enc = [values['valid']['lisi']['enc'][c][0].reshape(-1) for c in clusters]
        valid_lisi_rec = [values['valid']['lisi']['rec'][c][0].reshape(-1) for c in clusters]
        test_lisi_enc = [values['test']['lisi']['enc'][c][0].reshape(-1) for c in clusters]
        test_lisi_rec = [values['test']['lisi']['rec'][c][0].reshape(-1) for c in clusters]

        # set_lisi_enc = [values['set']['lisi']['enc'][c][0].reshape(-1) for c in ['labels', 'set']]
        # set_lisi_rec = [values['set']['lisi']['rec'][c][0].reshape(-1) for c in ['labels', 'set']]
    except:
        print("\n\n\nProblem with preparing LISI\n\n\n")
    try:
        lisi_train_reps = np.concatenate(
            [np.array([r for _ in np.concatenate(values)]) for r, values in zip(reps, [train_lisi_enc, train_lisi_rec])]
        )
        lisi_valid_reps = np.concatenate(
            [np.array([r for _ in np.concatenate(values)]) for r, values in zip(reps, [valid_lisi_enc, valid_lisi_rec])]
        )
        lisi_test_reps = np.concatenate(
            [np.array([r for _ in np.concatenate(values)]) for r, values in zip(reps, [test_lisi_enc, test_lisi_rec])]
        )
    except:
        print("\n\n\nProblem with concat LISI\n\n\n")
    try:
        lisi_df_train = pd.DataFrame(np.concatenate((
            np.concatenate((np.concatenate(train_lisi_enc), np.concatenate(train_lisi_rec))).reshape(1, -1),
            lisi_train_reps.reshape(1, -1),
            # np.concatenate((lisi_set_train_enc, lisi_set_train_rec)).reshape(1, -1),
        ), 0).T, columns=['lisi', 'representation'])
        lisi_df_train['lisi'] = pd.to_numeric(lisi_df_train['lisi'])
        lisi_df_valid = pd.DataFrame(np.concatenate((
            np.concatenate((valid_lisi_enc, valid_lisi_rec)).reshape(1, -1),
            lisi_valid_reps.reshape(1, -1),
            # np.concatenate((lisi_set_valid_enc, lisi_set_valid_rec)).reshape(1, -1),
        ), 0).T, columns=['lisi', 'representation'])
        lisi_df_valid['lisi'] = pd.to_numeric(lisi_df_valid['lisi'])
        lisi_df_test = pd.DataFrame(np.concatenate((
            np.concatenate((test_lisi_enc, test_lisi_rec)).reshape(1, -1),
            lisi_test_reps.reshape(1, -1),
            # np.concatenate((lisi_set_test_enc, lisi_set_test_rec)).reshape(1, -1),
        ), 0).T, columns=['lisi', 'representation'])
        lisi_df_test['lisi'] = pd.to_numeric(lisi_df_test['lisi'])
        # lisi_means = [values[s]['lisi'][0] for s in ['train', 'valid', 'test']]
    except:
        print("\n\n\nProblem with plotting LISI\n\n\n")

    try:
        sns.set_theme(style="whitegrid")

        if lisi_df_train.shape[0] > 0:
            figure = plt.figure(figsize=(8, 8))
            ax = sns.boxplot(x="representation", y="lisi", data=lisi_df_train)
            # logger.add_figure(f"LISI_train", figure, epoch)
            if mlops == "tensorboard":
                logger.add_figure(f"LISI_train", figure, epoch)
            elif mlops == "neptune":
                logger[f'LISI_train'].log(figure)
            elif mlops == "mlflow":
                mlflow.log_figure(figure, f'LISI_train.png')
            plt.close(figure)

        else:
            print('LISI_train Problem')

        if lisi_df_valid.shape[0] > 0:
            figure = plt.figure(figsize=(8, 8))
            ax = sns.boxplot(x="representation", y="lisi", data=lisi_df_valid)
            if mlops == "tensorboard":
                logger.add_figure(f"LISI_valid", figure, epoch)
            elif mlops == "neptune":
                logger[f'LISI_valid'].log(figure)
            elif mlops == "mlflow":
                mlflow.log_figure(figure, f'LISI_valid.png')
            plt.close(figure)

        else:
            print('LISI_valid Problem')
        if lisi_df_test.shape[0] > 0:
            figure = plt.figure(figsize=(8, 8))
            ax = sns.boxplot(x="representation", y="lisi", data=lisi_df_test)
            sns.set_theme(style="white")
            if mlops == "tensorboard":
                logger.add_figure(f"LISI_test", figure, epoch)
            elif mlops == "neptune":
                logger[f'LISI_test'].log(figure)
            elif mlops == "mlflow":
                mlflow.log_figure(figure, f'LISI_test.png')
            plt.close(figure)

        else:
            print('LISI_test Problem')
    except:
        print("\n\n\nProblem with plotting LISI\n\n\n")

    train_enc = torch.Tensor(np.concatenate(lists['train']['encoded_values']))
    valid_enc = torch.Tensor(np.concatenate(lists['valid']['encoded_values']))
    test_enc = torch.Tensor(np.concatenate(lists['test']['encoded_values']))
    if n_meta_emb > 0:
        train_enc = torch.cat((train_enc,
                               torch.Tensor(np.concatenate(lists['train']['age'])).view(-1, 1),
                               torch.Tensor(np.concatenate(lists['train']['gender'])).view(-1, 1),
                               ), 1)
        valid_enc = torch.cat((valid_enc,
                               torch.Tensor(np.concatenate(lists['valid']['age'])).view(-1, 1),
                               torch.Tensor(np.concatenate(lists['valid']['gender'])).view(-1, 1),
                               ), 1)
        test_enc = torch.cat((test_enc,
                              torch.Tensor(np.concatenate(lists['test']['age'])).view(-1, 1),
                              torch.Tensor(np.concatenate(lists['test']['gender'])).view(-1, 1),
                              ), 1)
    try:
        save_roc_curve(model,
                       train_enc.to(device),
                       np.concatenate(lists['train']['classes']),
                       unique_labels, name='./roc_train', binary=bout, epoch=epoch,
                       acc=values['train']['acc'][-1], logger=logger, mlops=mlops)
        save_roc_curve(model,
                       valid_enc.to(device),
                       np.concatenate(lists['valid']['classes']),
                       unique_labels, name='./roc_valid', binary=bout, epoch=epoch,
                       acc=values['valid']['acc'][-1], logger=logger, mlops=mlops)
        save_roc_curve(model,
                       test_enc.to(device),
                       np.concatenate(lists['test']['classes']),
                       unique_labels, name='./roc_test', binary=bout, epoch=epoch,
                       acc=values['test']['acc'][-1], logger=logger, mlops=mlops)
    except:
        print("\n\n\nProblem with ROC curves\n\n\n")

    try:
        save_precision_recall_curve(model,
                                    train_enc.to(device),
                                    np.concatenate(lists['train']['classes']),
                                    unique_labels, name='./prc_train', binary=bout, epoch=epoch,
                                    acc=values['train']['acc'][-1], logger=logger, mlops=mlops)
        save_precision_recall_curve(model,
                                    valid_enc.to(device),
                                    np.concatenate(lists['valid']['classes']),
                                    unique_labels, name='./prc_valid', binary=bout, epoch=epoch,
                                    acc=values['valid']['acc'][-1], logger=logger, mlops=mlops)
        save_precision_recall_curve(model,
                                    test_enc.to(device),
                                    np.concatenate(lists['test']['classes']),
                                    unique_labels, name='./prc_test', binary=bout, epoch=epoch,
                                    acc=values['test']['acc'][-1], logger=logger, mlops=mlops)
    except:
        print("\n\n\nProblem with precision/recall curves\n\n\n")

    return metrics


def make_data(lists, values):
    n_mini_batches = int(1000/lists['train'][values][0].shape[0])
    try:
        data = {
            'inputs': {group: np.concatenate(lists[group][values][:n_mini_batches]) for group in list(lists.keys()) if
                       len(lists[group][values]) > 0},
            'labels': {group: np.concatenate(lists[group]['labels'][:n_mini_batches]) for group in list(lists.keys()) if
                       len(lists[group][values]) > 0},
            'batches': {group: np.concatenate(lists[group]['domains'][:n_mini_batches]) for group in list(lists.keys()) if
                        len(lists[group][values]) > 0},
            # 'age': {group: np.concatenate(lists[group]['age']) for group in list(lists.keys()) if
            #         len(lists[group][values]) > 0},
            # 'gender': {group: np.concatenate(lists[group]['gender']) for group in list(lists.keys()) if
            #            len(lists[group][values]) > 0},
            # 'atn': {group: np.array(lists[group]['atn']) for group in list(lists.keys()) if
            #         len(lists[group][values]) > 0},
        }
        keys = list(data['inputs'].keys())
    except:
        data = {
            'inputs': {group: np.concatenate(lists[group][values][:n_mini_batches]) for group in ['all', 'all_pool']},
            'labels': {group: np.concatenate(lists[group]['labels'][:n_mini_batches]) for group in ['all', 'all_pool']},
            'batches': {group: np.concatenate(lists[group]['domains'][:n_mini_batches]) for group in ['all', 'all_pool']},
            # 'age': {group: np.concatenate(lists[group]['age']) for group in ['all', 'all_pool']},
            # 'gender': {group: np.concatenate(lists[group]['gender']) for group in ['all', 'all_pool']},
            # 'atn': {group: np.array(lists[group]['atn']) for group in ['all', 'all_pool']},
        }
        keys = ['all', 'all_pool']

    # for group in keys:
    #     meta_age = []
    #     for age in data['age'][group]:
    #         if age < 60:
    #             meta_age += ['50s']
    #         elif age < 70:
    #             meta_age += ['60s']
    #         elif age < 80:
    #             meta_age += ['70s']
    #         else:
    #             meta_age += ['80+']
    #     data['age'][group] = np.array(meta_age)

    return data


def log_plots(logger, lists, mlops, epoch):
    try:
        for name, values in zip(['encs', 'inputs', 'recs'], ['encoded_values', 'inputs', 'rec_values']):
            unique_labels = get_unique_labels(np.concatenate(lists['train']['labels']))
            unique_batches = np.unique(
                np.concatenate((
                    np.unique(np.concatenate(lists['train']['domains'])),
                    np.unique(np.concatenate(lists['valid']['domains'])),
                    np.unique(np.concatenate(lists['test']['domains'])),

                ))
            )
            # unique_ages = np.array(['50s', '60s', '70s', '80+'])
            # unique_genders = np.unique(np.concatenate(lists['all']['gender']))
            # unique_atns = np.unique([str(x) for x in np.array(lists['all']['atn'])])
            if len(lists['test'][values]) == 0:
                continue
            uniques = {'batches': unique_batches, 'labels': unique_labels}
            try:
                log_CCA({'model': CCA(n_components=2), 'name': f'CCA_{name}'},
                        logger, make_data(lists, values), uniques, mlops, epoch)
            except:
                pass

            try:
                log_LDA({'model': LDA, 'name': f'LDA_{name}'},
                        logger, make_data(lists, values), uniques, mlops, epoch)
            except:
                pass
            log_ORD({'model': PCA(n_components=2), 'name': f'PCA_{name}'}, logger,
                    make_data(lists, values), uniques, mlops, epoch),
            log_ORD({'model': PCA(n_components=2), 'name': f'PCA_{name}'}, logger,
                    make_data(lists, values), uniques, mlops, epoch, transductive=True),
            log_ORD({'model': UMAP(n_neighbors=5, min_dist=0.3, metric='correlation'), 'name': f'UMAP_{name}'},
                    logger, make_data(lists, values), uniques, mlops, epoch)
            log_ORD({'model': UMAP(n_neighbors=5, min_dist=0.3, metric='correlation'), 'name': f'UMAP_{name}'},
                    logger, make_data(lists, values), uniques, mlops, epoch, transductive=True)
    except:
        print("\n\n\nProblem with logging PCA, TSNE or UMAP\n\n\n")


def log_input_ordination(logger, data, scaler, mlops, epoch=0):
    try:
        data['gender'] = {}
        data['age'] = {}
        data['atn'] = {}
        for group in ['train', 'valid', 'test']:
            tmp = scaler.inverse_transform(data['inputs'][group])
            data['gender'][group] = tmp[:, -1].astype(np.int)
            data['age'][group] = tmp[:, -2]
            data['atn'][group] = tmp[:, -5:-2]
            data['atn'][group] = np.array([str(x) for x in data['atn'][group]])
            meta_age = []
            for age in data['age'][group]:
                if age < 60:
                    meta_age += ['50s']
                elif age < 70:
                    meta_age += ['60s']
                elif age < 80:
                    meta_age += ['70s']
                else:
                    meta_age += ['80+']
            data['age'][group] = np.array(meta_age)

        unique_labels = get_unique_labels(data['labels']['train'])
        unique_batches = np.unique(np.concatenate([
            data['batches']['train'],
            data['batches']['valid'],
            data['batches']['test']
        ]))
        unique_ages = np.array(['50s', '60s', '70s', '80+'])
        unique_genders = np.unique(data['gender']['train'])
        unique_atns = np.unique([str(x) for x in data['atn']['train']])
        uniques = {'batches': unique_batches, 'labels': unique_labels, 'age': unique_ages,
                   'gender': unique_genders, 'atn': unique_atns}
        try:
            log_CCA({'model': CCA(n_components=2), 'name': f'CCA_inputs'},
                    logger, data, uniques, mlops, epoch)
        except:
            pass
        try:
            log_LDA({'model': LDA, 'name': f'LDA_inputs'},
                    logger, data, uniques, mlops, epoch)
        except:
            pass
        log_ORD({'model': PCA(n_components=2), 'name': f'PCA_inputs'}, logger,
                data, uniques, mlops, epoch),
        log_ORD({'model': PCA(n_components=2), 'name': f'PCA_inputs'}, logger,
                data, uniques, mlops, epoch),
        log_ORD({'model': PCA(n_components=2), 'name': f'PCA_inputs'}, logger,
                data, uniques, mlops, epoch, transductive=True),
        log_ORD({'model': UMAP(n_neighbors=5, min_dist=0.3, metric='correlation'), 'name': f'UMAP_inputs'},
                logger, data, mlops, uniques, epoch)
        log_ORD({'model': UMAP(n_neighbors=5, min_dist=0.3, metric='correlation'), 'name': f'UMAP_inputs'},
                logger, data, mlops, uniques, epoch, transductive=True)

    except:
        print("\n\n\nProblem with logging PCA, TSNE or UMAP\n\n\n")



NEPTUNE_API_TOKEN = "YOUR-API-KEY"
NEPTUNE_PROJECT_NAME = "YOUR-PROJECT-NAME"
NEPTUNE_MODEL_NAME = "YOUR-MODEL-NAME"


import matplotlib
from bernn.utils.pool_metrics import log_pool_metrics

matplotlib.use('Agg')
CUDA_VISIBLE_DEVICES = ""

import uuid
import shutil

import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import random
import json
import copy
import torch
from itertools import cycle
from torch import nn
import os

from sklearn import metrics
from tensorboardX import SummaryWriter
from ax.service.managed_loop import optimize
from sklearn.metrics import matthews_corrcoef as MCC
from tensorboard.backend.event_processing.event_accumulator import EventAccumulator
from bernn.ml.train.params_gp import *
from bernn.utils.data_getters import get_alzheimer, get_amide, get_prostate, get_mice, get_data, get_bacteria
from bernn.dl.models.pytorch.aedann import ReverseLayerF
from bernn.dl.models.pytorch.aedann import AutoEncoder2 as AutoEncoder
from bernn.dl.models.pytorch.aedann import SHAPAutoEncoder2 as SHAPAutoEncoder
from bernn.dl.models.pytorch.utils.loggings import TensorboardLoggingAE, log_metrics, log_input_ordination, \
    LogConfusionMatrix, log_plots, log_neptune, log_shap, log_mlflow
from bernn.dl.models.pytorch.utils.dataset import get_loaders, get_loaders_no_pool
from bernn.utils.utils import scale_data, to_csv
from bernn.dl.models.pytorch.utils.utils import get_optimizer, to_categorical, get_empty_dicts, get_empty_traces, \
    log_traces, get_best_values, add_to_logger, add_to_neptune, add_to_mlflow
from bernn.dl.models.pytorch.utils.loggings import make_data
import neptune.new as neptune
import mlflow
import warnings
from datetime import datetime

warnings.filterwarnings("ignore")

random.seed(1)
torch.manual_seed(1)
np.random.seed(1)

class TrainAE:

    def __init__(self, args, path, fix_thres=-1, load_tb=False, log_metrics=False, keep_models=True, log_inputs=True,
                 log_plots=False, log_tb=False, log_neptune=False, log_mlflow=True, groupkfold=True, pools=True):
        """

        Args:
            args: contains multiple arguments passed in the command line
            log_path: Path where the tensorboard logs are saved
            path: Path to the data (in .csv format)
            fix_thres: If 1 > fix_thres >= 0 then the threshold is fixed to that value.
                       any other value means the threshold won't be fixed and will be
                       learned as an hyperparameter
            load_tb: If True, loads previous runs already saved
        """
        self.hparams_names = None
        self.best_acc = 0
        self.best_mcc = -1
        self.best_closs = np.inf
        self.logged_inputs = False
        self.log_tb = log_tb
        self.log_neptune = log_neptune
        self.log_mlflow = log_mlflow
        self.args = args
        self.path = path
        self.log_metrics = log_metrics
        self.log_plots = log_plots
        self.log_inputs = log_inputs
        self.keep_models = keep_models
        self.fix_thres = fix_thres
        self.load_tb = load_tb
        self.groupkfold = groupkfold
        self.foldername = None

        self.verbose = 1

        self.n_cats = None
        self.data = None
        self.unique_labels = None
        self.unique_batches = None

        self.pools = True

    def make_samples_weights(self):
        self.n_batches = len(set(self.data['batches']['all']))
        self.class_weights = {
            label: 1 / (len(np.where(label == self.data['labels']['train'])[0]) /
                        self.data['labels']['train'].shape[0])
            for label in self.unique_labels if
            label in self.data['labels']['train'] and label not in ["MCI-AD", 'MCI-other', 'DEM-other', 'NPH']}
        self.unique_unique_labels = list(self.class_weights.keys())
        for group in ['train', 'valid', 'test']:
            inds_to_keep = np.array([i for i, x in enumerate(self.data['labels'][group]) if x in self.unique_labels])
            self.data['inputs'][group] = self.data['inputs'][group].iloc[inds_to_keep]
            try:
                self.data['names'][group] = self.data['names'][group].iloc[inds_to_keep]
            except:
                self.data['names'][group] = self.data['names'][group][inds_to_keep]

            self.data['labels'][group] = self.data['labels'][group][inds_to_keep]
            self.data['cats'][group] = self.data['cats'][group][inds_to_keep]
            self.data['batches'][group] = self.data['batches'][group][inds_to_keep]

        self.samples_weights = {
            group: [self.class_weights[label] if label not in ["MCI-AD", 'MCI-other', 'DEM-other', 'NPH'] else 0 for
                    name, label in
                    zip(self.data['names'][group],
                        self.data['labels'][group])] if group == 'train' else [
                1 if label not in ["MCI-AD", 'MCI-other', 'DEM-other', 'NPH'] else 0 for name, label in
                zip(self.data['names'][group], self.data['labels'][group])] for group in
            ['train', 'valid', 'test']}
        self.n_cats = len(self.unique_labels)  # + 1  # for pool samples
        self.scaler = None

    def train(self, params):
        """

        Args:
            params: Contains the hyperparameters to be optimized

        Returns:
            best_closs: The best classification loss on the valid set

        """
        start_time = datetime.now()
        # Fixing the hyperparameters that are not optimized
        if args.dloss not in ['revTriplet', 'revDANN', 'DANN',
                                                          'inverseTriplet', 'normae'] or 'gamma' not in params:
            # gamma = 0 will ensure DANN is not learned
            params['gamma'] = 0
        if not self.args.variational or 'beta' not in params:
            # beta = 0 because useless outside a variational autoencoder
            params['beta'] = 0
        if not self.args.zinb or 'zeta' not in params:
            # zeta = 0 because useless outside a zinb autoencoder
            params['zeta'] = 0
        if 1 > self.fix_thres >= 0:
            # fixes the threshold of 0s tolerated for a feature
            params['thres'] = self.fix_thres
        else:
            params['thres'] = 0

        # params['dropout'] = 0
        # params['smoothing'] = 0
        # params['margin'] = 0
        # params['wd'] = 0
        print(params)

        # Assigns the hyperparameters getting optimized
        smooth = params['smoothing']
        layer1 = params['layer1']
        layer2 = params['layer2']
        scale = params['scaler']
        gamma = params['gamma']
        beta = params['beta']
        zeta = params['zeta']
        thres = params['thres']
        wd = params['wd']
        nu = params['nu']
        lr = params['lr']

        dropout = params['dropout']
        margin = params['margin']

        self.args.scaler = scale
        self.args.warmup = params['warmup']
        self.args.ncols = params['ncols']
        self.args.n_features = -1
        # self.args.disc_b_warmup = params['disc_b_warmup']

        optimizer_type = 'adam'
        metrics = {'pool_metrics': {}}
        # self.log_path is where tensorboard logs are saved
        self.foldername = str(uuid.uuid4())

        self.complete_log_path = f'logs/ae_then_classifier_holdout/{self.foldername}'
        loggers = {'cm_logger': LogConfusionMatrix(self.complete_log_path)}
        print(f'See results using: tensorboard --logdir={self.complete_log_path} --port=6006')

        hparams_filepath = self.complete_log_path + '/hp'
        os.makedirs(hparams_filepath, exist_ok=True)
        self.args.model_name = 'ae_then_classifier'
        if self.log_tb:
            loggers['tb_logging'] = TensorboardLoggingAE(hparams_filepath, params, variational=self.args.variational,
                                                         zinb=self.args.zinb,
                                                         tw=self.args.tied_weights,
                                                         dloss=self.args.dloss,
                                                         tl=0, # to remove
                                                         pseudo=self.args.predict_tests,
                                                         train_after_warmup=self.args.train_after_warmup,
                                                         berm='no', # to remove
                                                         args=self.args)
        if self.log_neptune:
            # Create a Neptune run object
            run = neptune.init_run(
                project=NEPTUNE_PROJECT_NAME,
                api_token=NEPTUNE_API_TOKEN,
            )  # your credentials
            model = neptune.init_model_version(
                model=NEPTUNE_MODEL_NAME,
                project=NEPTUNE_PROJECT_NAME,
                api_token=NEPTUNE_API_TOKEN,
                # your credentials
            )
            run["dataset"].track_files(f"{self.path}/{self.args.csv_file}")
            run["metadata"].track_files(
                f"{self.path}/subjects_experiment_ATN_verified_diagnosis.csv"
            )
            # Track metadata and hyperparameters by assigning them to the run
            model["inputs_type"] = run["inputs_type"] = args.csv_file.split(".csv")[0]
            model["best_unique"] = run["best_unique"] = args.best_features_file.split(".tsv")[0]
            model["use_valid"] = run["use_valid"] = args.use_valid
            model["use_test"] = run["use_test"] = args.use_test
            model["tied_weights"] = run["tied_weights"] = args.tied_weights
            model["random_recs"] = run["random_recs"] = args.random_recs
            model["train_after_warmup"] = run["train_after_warmup"] = args.train_after_warmup
            model["triplet_loss"] = run["triplet_loss"] = args.triplet_loss
            model["dloss"] = run["dloss"] = args.dloss
            model["predict_tests"] = run["predict_tests"] = args.predict_tests
            model["variational"] = run["variational"] = args.variational
            model["zinb"] = run["zinb"] = args.zinb
            model["threshold"] = run["threshold"] = args.threshold
            model["rec_loss_type"] = run["rec_loss_type"] = args.rec_loss
            model["strategy"] = run["strategy"] = args.strategy
            model["bad_batches"] = run["bad_batches"] = args.bad_batches
            model["remove_zeros"] = run["remove_zeros"] = args.remove_zeros
            model["parameters"] = run["parameters"] = params
            model["csv_file"] = run["csv_file"] = args.csv_file
            model["model_name"] = run["model_name"] = 'ae_then_classifier'
            model["n_meta"] = run["n_meta"] = args.n_meta
            model["n_emb"] = run["n_emb"] = args.embeddings_meta
            model["groupkfold"] = run["groupkfold"] = args.groupkfold
            model["embeddings_meta"] = run["embeddings_meta"] = args.embeddings_meta
            model["foldername"] = run["foldername"] = self.foldername
            model["use_mapping"] = run["use_mapping"] = args.use_mapping
            model["dataset_name"] = run["dataset_name"] = args.dataset
            model["n_agg"] = run["n_agg"] = args.n_agg
        else:
            model = None
            run = None

        if self.log_mlflow:
            mlflow.set_experiment(
                self.args.exp_id,
            )
            try:
                mlflow.start_run()
            except:
                mlflow.end_run()
                mlflow.start_run()
            mlflow.log_params({
                "inputs_type": args.csv_file.split(".csv")[0],
                "best_unique": args.best_features_file.split(".tsv")[0],
                "tied_weights": args.tied_weights,
                "random_recs": args.random_recs,
                "train_after_warmup": args.train_after_warmup,
                "dloss": args.dloss,
                "predict_tests": args.predict_tests,
                "variational": args.variational,
                "zinb": args.zinb,
                "threshold": args.threshold,
                "rec_loss_type": args.rec_loss,
                "bad_batches": args.bad_batches,
                "remove_zeros": args.remove_zeros,
                "parameters": params,
                "scaler": params['scaler'],
                "csv_file": args.csv_file,
                "model_name": args.model_name,
                "n_meta": args.n_meta,
                "n_emb": args.embeddings_meta,
                "groupkfold": args.groupkfold,
                "n_features": self.args.n_features,
                "foldername": self.foldername,
                "use_mapping": args.use_mapping,
                "dataset_name": args.dataset,
                "n_agg": args.n_agg,
            })
        else:
            model = None
            run = None
        seed = 0
        combinations = []
        h = 0
        best_closses = []
        best_mccs = []

        # warmup is done only once, at first repeat
        warmup_counter = 0
        warmup_b_counter = 0
        if self.args.warmup > 0:
            warmup = True
        else:
            warmup = False
        warmup_disc_b = False
        epoch = 0
        best_loss = np.inf
        best_closs = np.inf
        best_dom_loss = np.inf
        best_dom_acc = np.inf
        best_acc = 0
        best_mcc = -np.inf
        if self.args.dataset == 'alzheimer':
            self.data, self.unique_labels, self.unique_batches = get_alzheimer(self.path, args, seed=seed)
            self.pools = True
        elif self.args.dataset == 'amide':
            self.data, self.unique_labels, self.unique_batches = get_amide(self.path, args, seed=seed)
            self.pools = True

        elif self.args.dataset == 'bacteria':
            self.data, self.unique_labels, self.unique_batches = get_bacteria(self.path, args, seed=seed)
            self.pools = False

        elif self.args.dataset == 'prostate':
            self.data, self.unique_labels, self.unique_batches = get_prostate(self.path, args, seed=seed)

        elif self.args.dataset == 'mice':
            self.data, self.unique_labels, self.unique_batches = get_mice(self.path, args, seed=seed)
        else:
            exit('Wrong dataset name')
        # self.get_amide(self.path, seed=(1 + h) * 10)
        # print(combinations)
        self.columns = self.data['inputs']['all'].columns
        self.make_samples_weights()
        # event_acc is used to verify if the hparams have already been tested. If they were,
        # the best classification loss is retrieved and we go to the next trial
        event_acc = EventAccumulator(hparams_filepath)
        event_acc.Reload()
        # Transform the data with the chosen scaler
        data = copy.deepcopy(self.data)
        data, self.scaler = scale_data(scale, data, self.args.device)

        for g in list(data['inputs'].keys()):
            data['inputs'][g] = data['inputs'][g].round(4)
        # Gets all the pytorch dataloaders to train the models
        if self.pools:
            loaders = get_loaders(data, self.args.random_recs, self.samples_weights, self.args.dloss, None,
                                  None, bs=64)
        else:
            loaders = get_loaders_no_pool(data, self.args.random_recs, self.samples_weights, self.args.dloss,
                                          None, None, bs=8)

        ae = AutoEncoder(data['inputs']['all'].shape[1],
                         n_batches=self.n_batches,
                         nb_classes=self.n_cats,
                         mapper=self.args.use_mapping,
                         layer1=layer1,
                         layer2=layer2,
                         n_layers=self.args.n_layers,
                         n_meta=self.args.n_meta,
                         n_emb=self.args.embeddings_meta,
                         dropout=dropout,
                         variational=self.args.variational, conditional=False,
                         zinb=self.args.zinb, add_noise=0, tied_weights=self.args.tied_weights,
                         use_gnn=self.args.use_gnn, device=self.args.device).to(self.args.device)
        ae.mapper.to(self.args.device)
        ae.dec.to(self.args.device)
        # if self.args.embeddings_meta > 0:
        #     n_meta = self.n_meta
        shap_ae = SHAPAutoEncoder(data['inputs']['all'].shape[1],
                                  n_batches=self.n_batches,
                                  nb_classes=self.n_cats,
                                  mapper=self.args.use_mapping,
                                  layer1=layer1,
                                  layer2=layer2,
                                  n_layers=self.args.n_layers,
                                  n_meta=self.args.n_meta,
                                  n_emb=self.args.embeddings_meta,
                                  dropout=dropout,
                                  variational=self.args.variational, conditional=False,
                                  zinb=self.args.zinb, add_noise=0, tied_weights=self.args.tied_weights,
                                  use_gnn=self.args.use_gnn, device=self.args.device).to(self.args.device)
        shap_ae.mapper.to(self.args.device)
        shap_ae.dec.to(self.args.device)
        loggers['logger_cm'] = SummaryWriter(f'{self.complete_log_path}/cm')
        loggers['logger'] = SummaryWriter(f'{self.complete_log_path}/traces')
        sceloss, celoss, mseloss, triplet_loss = self.get_losses(scale, smooth, margin, args.dloss)

        optimizer_ae = get_optimizer(ae, lr, wd, optimizer_type)
        self.optimizer_c = get_optimizer(ae.classifier, nu * lr, wd, optimizer_type)

        # Used only if bdisc==1
        optimizer_b = get_optimizer(ae.dann_discriminator, 1e-2, 0, optimizer_type)

        self.hparams_names = [x.name for x in linsvc_space]
        if self.log_inputs and not self.logged_inputs:
            data['inputs']['all'].to_csv(
                f'{self.complete_log_path}/{self.args.berm}_inputs.csv')
            if self.log_neptune:
                run[f"inputs.csv"].track_files(f'{self.complete_log_path}/{self.args.berm}_inputs.csv')
            log_input_ordination(loggers['logger'], data, self.scaler, epoch)
            if self.pools:
                metrics = log_pool_metrics(data['inputs'], data['batches'], data['labels'],
                                           self.unique_unique_labels, loggers, epoch, metrics, 'inputs')
            self.logged_inputs = True

        values, best_values, _, best_traces = get_empty_dicts()

        early_stop_counter = 0
        best_vals = values
        if h > 1:  # or warmup_counter == 100:
            ae.load_state_dict(torch.load(f'{self.complete_log_path}/warmup.pth'))
            print(f"\n\nNO WARMUP\n\n")
        # while new_combinations:
        for epoch in range(0, self.args.warmup):
            lists, traces = get_empty_traces()
            ae.train()

            iterator = enumerate(loaders['all'])

            # If option train_after_warmup=1, then this loop is only for preprocessing
            if warmup or self.args.train_after_warmup:
                for i, all_batch in iterator:
                    if warmup or self.args.train_after_warmup:
                        optimizer_ae.zero_grad()
                    inputs, meta_inputs, names, labels, domain, to_rec, not_to_rec, pos_batch_sample, \
                        neg_batch_sample, meta_pos_batch_sample, meta_neg_batch_sample, set = all_batch
                    inputs = inputs.to(self.args.device).float()
                    meta_inputs = meta_inputs.to(self.args.device).float()
                    to_rec = to_rec.to(self.args.device).float()
                    if self.args.n_meta > 0:
                        inputs = torch.cat((inputs, meta_inputs), 1)
                        to_rec = torch.cat((to_rec, meta_inputs), 1)

                    enc, rec, zinb_loss, kld = ae(inputs, to_rec, domain, sampling=True)
                    rec = rec['mean']
                    zinb_loss = zinb_loss.to(self.args.device)
                    reverse = ReverseLayerF.apply(enc, 1)
                    if args.dloss == 'DANN':
                        domain_preds = ae.dann_discriminator(reverse)
                    else:
                        domain_preds = ae.dann_discriminator(enc)
                    if args.dloss not in ['revTriplet', 'inverseTriplet']:
                        dloss, domain = self.get_dloss(celoss, domain, domain_preds, 2)
                    elif args.dloss == 'revTriplet':
                        pos_batch_sample = pos_batch_sample.to(self.args.device).float()
                        neg_batch_sample = neg_batch_sample.to(self.args.device).float()
                        meta_pos_batch_sample = meta_pos_batch_sample.to(self.args.device).float()
                        meta_neg_batch_sample = meta_neg_batch_sample.to(self.args.device).float()
                        if self.args.n_meta > 0:
                            pos_batch_sample = torch.cat((pos_batch_sample, meta_pos_batch_sample), 1)
                            neg_batch_sample = torch.cat((neg_batch_sample, meta_neg_batch_sample), 1)
                        pos_enc, _, _, _ = ae(pos_batch_sample, pos_batch_sample, domain, sampling=True)
                        neg_enc, _, _, _ = ae(neg_batch_sample, neg_batch_sample, domain, sampling=True)
                        dloss = triplet_loss(reverse,
                                             ReverseLayerF.apply(pos_enc, 1),
                                             ReverseLayerF.apply(neg_enc, 1)
                                             )
                    elif args.dloss == 'inverseTriplet':
                        pos_batch_sample, neg_batch_sample = neg_batch_sample.to(
                            self.args.device).float(), pos_batch_sample.to(self.args.device).float()
                        meta_pos_batch_sample, meta_neg_batch_sample = meta_neg_batch_sample.to(
                            self.args.device).float(), meta_pos_batch_sample.to(self.args.device).float()
                        if self.args.n_meta > 0:
                            pos_batch_sample = torch.cat((pos_batch_sample, meta_pos_batch_sample), 1)
                            neg_batch_sample = torch.cat((neg_batch_sample, meta_neg_batch_sample), 1)
                        pos_enc, _, _, _ = ae(pos_batch_sample, pos_batch_sample, domain, sampling=True)
                        neg_enc, _, _, _ = ae(neg_batch_sample, neg_batch_sample, domain, sampling=True)
                        dloss = triplet_loss(enc, pos_enc, neg_enc)
                        # domain = domain.argmax(1)

                    if torch.isnan(enc[0][0]):
                        if self.log_mlflow:
                            mlflow.log_param('finished', 0)
                            mlflow.end_run()
                        return best_loss

                    if isinstance(rec, list):
                        rec = rec[-1]
                    if isinstance(to_rec, list):
                        to_rec = to_rec[-1]
                    if scale == 'binarize':
                        rec = torch.sigmoid(rec)
                    rec_loss = mseloss(rec, to_rec)
                    if zinb_loss > 0:
                        rec_loss = zinb_loss
                    # else:
                    #     rec_loss = zinb_loss
                    traces['rec_loss'] += [rec_loss.item()]
                    traces['dom_loss'] += [dloss.item()]
                    traces['dom_acc'] += [np.mean([0 if pred != dom else 1 for pred, dom in
                                                   zip(domain_preds.detach().cpu().numpy().argmax(1),
                                                       domain.detach().cpu().numpy())])]
                    # lists['all']['set'] += [np.array([group for _ in range(len(domain))])]
                    lists['all']['domains'] += [np.array(
                        [self.unique_batches[d] for d in domain.detach().cpu().numpy()])]
                    lists['all']['domain_preds'] += [domain_preds.detach().cpu().numpy()]
                    # lists[group]['preds'] += [preds.detach().cpu().numpy()]
                    lists['all']['classes'] += [labels.detach().cpu().numpy()]
                    lists['all']['encoded_values'] += [
                        enc.detach().cpu().numpy()]
                    lists['all']['rec_values'] += [
                        rec.detach().cpu().numpy()]
                    lists['all']['names'] += [names]
                    lists['all']['gender'] += [meta_inputs.detach().cpu().numpy()[:, -1]]
                    lists['all']['age'] += [meta_inputs.detach().cpu().numpy()[:, -2]]
                    lists['all']['atn'] += [str(x) for x in
                                            meta_inputs.detach().cpu().numpy()[:, -5:-2]]
                    lists['all']['inputs'] += [data['inputs']['all'].to_numpy()]
                    try:
                        lists['all']['labels'] += [np.array(
                            [self.unique_labels[x] for x in labels.detach().cpu().numpy()])]
                    except:
                        pass
                    if warmup or self.args.train_after_warmup and not warmup_disc_b:
                        (rec_loss + gamma * dloss + beta * kld.mean()).backward()
                        # (rec_loss + gamma * dloss + beta * kld.mean() + zeta * zinb_loss).backward()
                        nn.utils.clip_grad_norm_(ae.parameters(), max_norm=1)
                        optimizer_ae.step()
            else:
                ae = self.freeze_ae(ae)

            if np.mean(traces['rec_loss']) < best_loss:
                # "Every counters go to 0 when a better reconstruction loss is reached"
                print(
                    f"Best Loss Epoch {epoch}, Losses: {np.mean(traces['rec_loss'])}, "
                    f"Domain Losses: {np.mean(traces['dom_loss'])}, "
                    f"Domain Accuracy: {np.mean(traces['dom_acc'])}")
                warmup_counter = 0
                early_stop_counter = 0
                best_loss = np.mean(traces['rec_loss'])
                dom_loss = np.mean(traces['dom_loss'])
                dom_acc = np.mean(traces['dom_acc'])
                if warmup:
                    torch.save(ae.state_dict(), f'{self.complete_log_path}/warmup.pth')

            if (
                    self.args.early_warmup_stop != 0 and warmup_counter == self.args.early_warmup_stop) and warmup:  # or warmup_counter == 100:
                # When the warnup counter gets to
                values = log_traces(traces, values)
                if self.args.early_warmup_stop != 0:
                    try:
                        ae.load_state_dict(torch.load(f'{self.complete_log_path}/model.pth'))
                    except:
                        pass
                print(f"\n\nWARMUP FINISHED (early stop). {epoch}\n\n")
                warmup = False
                warmup_disc_b = True

            if epoch == self.args.warmup and warmup:  # or warmup_counter == 100:
                # When the warnup counter gets to
                if self.args.early_warmup_stop != 0:
                    try:
                        ae.load_state_dict(torch.load(f'{self.complete_log_path}/model.pth'))
                    except:
                        pass
                print(f"\n\nWARMUP FINISHED. {epoch}\n\n")
                values = log_traces(traces, values)
                warmup = False
                warmup_disc_b = True

            if epoch < self.args.warmup and warmup:  # and np.mean(traces['rec_loss']) >= best_loss:
                values = log_traces(traces, values)
                warmup_counter += 1
                # best_values = get_best_values(traces, ae_only=True)
                if self.log_tb:
                    loggers['tb_logging'].logging(values, metrics)
                if self.log_neptune:
                    log_neptune(run, values)
                if self.log_mlflow:
                    add_to_mlflow(values, epoch)
                continue
            ae.train()
            if self.args.bdisc:
                self.forward_discriminate(optimizer_b, ae, celoss, loaders['all'])
            if warmup_disc_b and warmup_b_counter < 0:
                warmup_b_counter += 1
                continue
            else:
                warmup_disc_b = False

        # If training of the autoencoder is retricted to the warmup, (train_after_warmup=0),
        # all layers except the classification layers are frozen

        ae.eval()
        for epoch in range(0, self.args.n_epochs):
            if early_stop_counter >= self.args.early_stop:
                if self.verbose > 0:
                    print('EARLY STOPPING.', epoch)
                break
            lists, traces = get_empty_traces()
            closs, _, _ = self.loop('train', ae, sceloss, loaders['train'], lists, traces, nu=nu)

            if torch.isnan(closs):
                if self.log_mlflow:
                    mlflow.log_param('finished', 0)
                    mlflow.end_run()
                return best_loss

            # Below is the loop for all sets
            with torch.no_grad():
                for group in list(data['inputs'].keys()):
                    if group in ['all', 'all_pool']:
                        continue
                    closs, lists, traces = self.loop(group, ae, sceloss, loaders[group], lists, traces, nu=0)

            traces = self.get_mccs(lists, traces)
            values = log_traces(traces, values)
            if self.log_tb:
                try:
                    add_to_logger(values, loggers['logger'], epoch)
                except:
                    print("Problem with add_to_logger!")
            if self.log_neptune:
                add_to_neptune(values, run, epoch)
            if self.log_mlflow:
                add_to_mlflow(values, epoch)
            if np.mean(values['valid']['mcc'][-self.args.n_agg:]) > best_mcc and len(
                    values['valid']['mcc']) > self.args.n_agg:
                print(f"Best Classification Mcc Epoch {epoch}, "
                      f"Acc: test: {values['test']['acc'][-1]}, valid: {values['valid']['acc'][-1]}, train: {values['train']['acc'][-1]}"
                      f"Mcc: {values['test']['mcc'][-1]}"
                      f"Classification train loss: {values['train']['closs'][-1]},"
                      f" valid loss: {values['valid']['closs'][-1]},"
                      f" test loss: {values['test']['closs'][-1]}")
                best_mcc = np.mean(values['valid']['mcc'][-self.args.n_agg:])
                torch.save(ae.state_dict(), f'{self.complete_log_path}/model.pth')
                best_values = get_best_values(values.copy(), ae_only=False, n_agg=self.args.n_agg)
                best_vals = values.copy()
                best_vals['rec_loss'] = best_loss
                best_vals['dom_loss'] = best_dom_loss
                best_vals['dom_acc'] = best_dom_acc
                early_stop_counter = 0

            if values['valid']['acc'][-1] > best_acc:
                print(f"Best Classification Mcc Epoch {epoch}, "
                      f"Acc: test: {values['test']['acc'][-1]}, valid: {values['valid']['acc'][-1]}, train: {values['train']['acc'][-1]}"
                      f"Mcc: {values['test']['mcc'][-1]}"
                      f"Classification train loss: {values['train']['closs'][-1]},"
                      f" valid loss: {values['valid']['closs'][-1]},"
                      f" test loss: {values['test']['closs'][-1]}")

                best_acc = values['valid']['acc'][-1]
                early_stop_counter = 0

            if values['valid']['closs'][-1] < best_closs:
                print(f"Best Classification Mcc Epoch {epoch}, "
                      f"Acc: test: {values['test']['acc'][-1]}, valid: {values['valid']['acc'][-1]}, train: {values['train']['acc'][-1]}"
                      f"Mcc: {values['test']['mcc'][-1]}"
                      f"Classification train loss: {values['train']['closs'][-1]},"
                      f" valid loss: {values['valid']['closs'][-1]},"
                      f" test loss: {values['test']['closs'][-1]}")
                best_closs = values['valid']['closs'][-1]
                early_stop_counter = 0
            else:
                # if epoch > self.warmup:
                early_stop_counter += 1

            if self.args.predict_tests and (epoch % 10 == 0):
                loaders = get_loaders(self.data, data, self.args.random_recs, self.args.triplet_dloss, ae,
                                      ae.classifier)

        best_mccs += [best_mcc]

        best_lists, traces = get_empty_traces()
        # Loading best model that was saved during training
        ae.load_state_dict(torch.load(f'{self.complete_log_path}/model.pth'))
        # Need another model because the other cant be use to get shap values
        shap_ae.load_state_dict(torch.load(f'{self.complete_log_path}/model.pth'))
        # ae.load_state_dict(sd)
        ae.eval()
        shap_ae.eval()
        with torch.no_grad():
            for group in list(data['inputs'].keys()):
                # if group in ['all', 'all_pool']:
                #     continue
                closs, best_lists, traces = self.loop(group, ae, sceloss, loaders[group], best_lists, traces, nu=0,
                                                      mapping=False)  # -1
        if self.log_neptune:
            model["model"].upload(f'{self.complete_log_path}/model.pth')
            model["validation/closs"].log(best_closs)
        best_closses += [best_closs]
        # logs things in the background. This could be problematic if the logging takes more time than each iteration of repetitive holdout
        # daemon = Thread(target=self.log_rep, daemon=True, name='Monitor',
        #                 args=[best_lists, best_vals, best_values, traces, model, metrics, run, cm_logger, ae,
        #                       shap_ae, h,
        #                       epoch])
        # daemon.start()
        self.log_rep(best_lists, best_vals, best_values, traces, model, metrics, run, loggers, ae,
                     shap_ae, h, epoch)
        del ae, shap_ae

        # Logging every model is taking too much resources and it makes it quite complicated to get information when
        # Too many runs have been made. This will make the notebook so much easier to work with
        if np.mean(best_mccs) > self.best_mcc:
            try:
                if os.path.exists(
                        f'logs/best_models/ae_then_classifier_holdout/{self.args.dataset}/{self.args.dloss}_vae{self.args.variational}'):
                    shutil.rmtree(
                        f'logs/best_models/ae_then_classifier_holdout/{self.args.dataset}/{self.args.dloss}_vae{self.args.variational}',
                        ignore_errors=True)
                shutil.copytree(f'{self.complete_log_path}',
                                f'logs/best_models/ae_then_classifier_holdout/{self.args.dataset}/{self.args.dloss}_vae{self.args.variational}')
                # print("File copied successfully.")

            # If source and destination are same
            except shutil.SameFileError:
                # print("Source and destination represents the same file.")
                pass
            self.best_mcc = np.mean(best_mccs)

        # Logs confusion matrices in the background. Also runs RandomForestClassifier on encoded and reconstructed
        # representations. This should be shorter than the actual calculation of the model above in the function,
        # otherwise the number of threads will keep increasing.
        # daemon = Thread(target=self.logging, daemon=True, name='Monitor', args=[run, cm_logger])
        # daemon.start()
        if self.log_mlflow:
            mlflow.log_param('finished', 1)
        self.logging(run, loggers['cm_logger'])

        if not self.keep_models:
            # shutil.rmtree(f'{self.complete_log_path}/traces', ignore_errors=True)
            # shutil.rmtree(f'{self.complete_log_path}/cm', ignore_errors=True)
            # shutil.rmtree(f'{self.complete_log_path}/hp', ignore_errors=True)
            shutil.rmtree(f'{self.complete_log_path}', ignore_errors=True)
        print('Duration: {}'.format(datetime.now() - start_time))
        best_closs = np.mean(best_closses)
        if best_closs < self.best_closs:
            self.best_closs = best_closs
            print("Best closs!")

        # It should not be necessary. To remove once certain the "Too many files open" error is no longer a problem
        plt.close('all')

        return best_mcc

    def log_rep(self, best_lists, best_vals, best_values, traces, model, metrics, run, loggers, ae, shap_ae, h,
                epoch):
        # if run is None:
        #     return None
        best_traces = self.get_mccs(best_lists, traces)

        self.log_predictions(best_lists, run, h)

        if self.log_metrics:
            if self.log_tb:
                try:
                    # logger, lists, values, model, unique_labels, mlops, epoch, metrics, n_meta_emb=0, device='cuda'
                    metrics = log_metrics(loggers['logger'], best_lists, best_vals, ae,
                                        np.unique(np.concatenate(best_lists['train']['labels'])),
                                        np.unique(self.data['batches']), epoch, mlops="tensorboard",
                                        metrics=metrics, n_meta_emb=self.args.embeddings_meta,
                                        device=self.args.device)
                except BrokenPipeError:
                    print("\n\n\nProblem with logging stuff!\n\n\n")
            if self.log_neptune:
                try:
                    metrics = log_metrics(run, best_lists, best_vals, ae,
                                        np.unique(np.concatenate(best_lists['train']['labels'])),
                                        np.unique(self.data['batches']), epoch=epoch, mlops="neptune",
                                        metrics=metrics, n_meta_emb=self.args.embeddings_meta,
                                        device=self.args.device)
                except BrokenPipeError:
                    print("\n\n\nProblem with logging stuff!\n\n\n")
            if self.log_mlflow:
                try:
                    metrics = log_metrics(None, best_lists, best_vals, ae,
                                        np.unique(np.concatenate(best_lists['train']['labels'])),
                                        np.unique(self.data['batches']), epoch, mlops="mlflow",
                                        metrics=metrics, n_meta_emb=self.args.embeddings_meta,
                                        device=self.args.device)
                except BrokenPipeError:
                    print("\n\n\nProblem with logging stuff!\n\n\n")

        if self.log_metrics and self.pools:
            try:
                if self.log_neptune:
                    enc_data = make_data(best_lists, 'encoded_values')
                    metrics = log_pool_metrics(enc_data['inputs'], enc_data['batches'], enc_data['labels'],
                                               self.unique_unique_labels, run, epoch, metrics, 'enc', 'neptune')
                    rec_data = make_data(best_lists, 'rec_values')
                    metrics = log_pool_metrics(rec_data['inputs'], rec_data['batches'], rec_data['labels'],
                                               self.unique_unique_labels, run, epoch, metrics, 'rec', 'neptune')
                if self.log_mlflow:
                    enc_data = make_data(best_lists, 'encoded_values')
                    metrics = log_pool_metrics(enc_data['inputs'], enc_data['batches'], enc_data['labels'],
                                               self.unique_unique_labels, run, epoch, metrics, 'enc', 'mlflow')
                    rec_data = make_data(best_lists, 'rec_values')
                    metrics = log_pool_metrics(rec_data['inputs'], rec_data['batches'], rec_data['labels'],
                                               self.unique_unique_labels, run, epoch, metrics, 'rec', 'mlflow')
                if self.log_tb:
                    enc_data = make_data(best_lists, 'encoded_values')
                    metrics = log_pool_metrics(enc_data['inputs'], enc_data['batches'], enc_data['labels'],
                                               self.unique_unique_labels, loggers['logger'], epoch, metrics, 'enc',
                                               'tensorboard')
                    rec_data = make_data(best_lists, 'rec_values')
                    metrics = log_pool_metrics(rec_data['inputs'], rec_data['batches'], rec_data['labels'],
                                               self.unique_unique_labels, loggers['logger'], epoch, metrics, 'rec',
                                               'tensorboard')

            except BrokenPipeError:
                print("\n\n\nProblem with logging stuff!\n\n\n")

        loggers['cm_logger'].add(best_lists)

        if self.log_plots:
            if self.log_tb:
                # TODO Add log_shap
                # logger.add(loggers['logger_cm'], epoch, best_lists,
                #            self.unique_labels, best_traces, 'tensorboard')
                log_plots(loggers['logger_cm'], best_lists, 'tensorboard', epoch)
                log_shap(loggers['logger_cm'], shap_ae, best_lists, self.columns, self.args.embeddings_meta, 'tb',
                         self.complete_log_path,
                         self.args.device)
            if self.log_neptune:
                log_shap(run, shap_ae, best_lists, self.columns, self.args.embeddings_meta, 'neptune', self.complete_log_path,
                         self.args.device)
                log_plots(run, best_lists, 'neptune', epoch)
            if self.log_mlflow:
                log_shap(None, shap_ae, best_lists, self.columns, self.args.embeddings_meta, 'mlflow',
                         self.complete_log_path,
                         self.args.device)
                log_plots(None, best_lists, 'mlflow', epoch)

        columns = self.data['inputs']['all'].columns
        if self.args.n_meta == 2:
            columns += ['gender', 'age']

        rec_data, enc_data = to_csv(best_lists, self.complete_log_path, self.data['inputs']['all'].columns)

        if self.log_neptune:
            run["recs"].track_files(f'{self.complete_log_path}/recs.csv')
            run["encs"].track_files(f'{self.complete_log_path}/encs.csv')

        best_values['pool_metrics'] = {}
        try:
            best_values['batches'] = metrics['batches']
        except:
            pass
        try:
            best_values['pool_metrics']['enc'] = metrics['pool_metrics_enc']
        except:
            pass
        try:
            best_values['pool_metrics']['rec'] = metrics['pool_metrics_rec']
        except:
            pass

        if self.log_tb:
            loggers['tb_logging'].logging(best_values, metrics)
        if self.log_neptune:
            log_neptune(run, best_values)
        if self.log_mlflow:
            log_mlflow(best_values, h)

        # except BrokenPipeError:
        #     print("\n\n\nProblem with logging stuff!\n\n\n")

    def logging(self, run, cm_logger):

        if self.log_neptune:
            cm_logger.plot(run, 0, self.unique_unique_labels, 'neptune')
            # cm_logger.get_rf_results(run, self.args)
            run.stop()
        if self.log_mlflow:
            cm_logger.plot(None, 0, self.unique_unique_labels, 'mlflow')
            # cm_logger.get_rf_results(run, self.args)
            mlflow.end_run()
        # cm_logger.close()
        # logger.close()

    def log_predictions(self, best_lists, run, step):
        cats, labels, preds, scores, names = [{'train': [], 'valid': [], 'test': []} for _ in range(5)]
        for group in ['train', 'valid', 'test']:
            cats[group] = np.concatenate(best_lists[group]['cats'])
            labels[group] = np.concatenate(best_lists[group]['labels'])
            scores[group] = torch.softmax(torch.Tensor(np.concatenate(best_lists[group]['preds'])), 1)
            preds[group] = scores[group].argmax(1)
            names[group] = np.concatenate(best_lists[group]['names'])
            pd.DataFrame(np.concatenate((labels[group].reshape(-1, 1), scores[group],
                                         np.array([self.unique_labels[x] for x in preds[group]]).reshape(-1, 1),
                                         names[group].reshape(-1, 1)), 1)).to_csv(
                f'{self.complete_log_path}/{group}_predictions.csv')
            if self.log_neptune:
                try:
                    run[f"{group}_predictions"].track_files(f'{self.complete_log_path}/{group}_predictions.csv')
                    run[f'{group}_AUC'] = metrics.roc_auc_score(y_true=cats[group], y_score=scores[group],
                                                                multi_class='ovr')
                except:
                    pass
            if self.log_mlflow:
                try:
                    mlflow.log_metric(f'{group}_AUC',
                                  metrics.roc_auc_score(y_true=cats[group], y_score=scores[group], multi_class='ovr'),
                                  step=step)
                except:
                    pass

    def loop(self, group, ae, celoss, loader, lists, traces, nu=1, mapping=True):
        """

        Args:
            group: Which set? Train, valid or test
            optimizer_ae: Object that contains the optimizer for the autoencoder
            ae: AutoEncoder (pytorch model, inherits nn.Module)
            celoss: torch.nn.CrossEntropyLoss instance
            triplet_loss: torch.nn.TripletMarginLoss instance
            loader: torch.utils.data.DataLoader
            lists: List keeping informations on the current run
            traces: List keeping scores on the current run
            nu: hyperparameter controlling the importance of the classification loss

        Returns:

        """
        # If group is train and nu = 0, then it is not training. valid can also have sampling = True
        if group in ['train', 'valid'] and nu != 0:
            sampling = True
        else:
            sampling = False
        classif_loss = None
        for i, batch in enumerate(loader):
            if group in ['train'] and nu != 0:
                self.optimizer_c.zero_grad()
            data, meta_inputs, names, labels, domain, to_rec, not_to_rec, pos_batch_sample, \
                neg_batch_sample, meta_pos_batch_sample, meta_neg_batch_sample, set = batch
            # data[torch.isnan(data)] = 0
            data = data.to(self.args.device).float()
            meta_inputs = meta_inputs.to(self.args.device).float()
            to_rec = to_rec.to(self.args.device).float()

            # If n_meta > 0, meta data added to inputs
            if self.args.n_meta > 0:
                data = torch.cat((data, meta_inputs), 1)
                to_rec = torch.cat((to_rec, meta_inputs), 1)
            not_to_rec = not_to_rec.to(self.args.device).float()
            enc, rec, _, kld = ae(data, to_rec, domain, sampling=sampling, mapping=mapping)
            rec = rec['mean']

            # If embedding_meta > 0, meta data added to embeddings
            if self.args.embeddings_meta:
                preds = ae.classifier(torch.cat((enc, meta_inputs), 1))
            else:
                preds = ae.classifier(enc)

            domain_preds = ae.dann_discriminator(enc)
            try:
                cats = to_categorical(labels.long(), self.n_cats).to(self.args.device).float()
                classif_loss = celoss(preds, cats)
            except:
                cats = torch.Tensor([self.n_cats + 1 for _ in labels])
                classif_loss = torch.Tensor([0])

            if not self.args.zinb:
                if isinstance(rec, list):
                    rec = rec[-1]
                if isinstance(to_rec, list):
                    to_rec = to_rec[-1]
            lists[group]['set'] += [np.array([group for _ in range(len(domain))])]
            lists[group]['domains'] += [np.array([self.unique_batches[d] for d in domain.detach().cpu().numpy()])]
            lists[group]['domain_preds'] += [domain_preds.detach().cpu().numpy()]
            lists[group]['preds'] += [preds.detach().cpu().numpy()]
            lists[group]['classes'] += [labels.detach().cpu().numpy()]
            # lists[group]['encoded_values'] += [enc.view(enc.shape[0], -1).detach().cpu().numpy()]
            lists[group]['names'] += [names]
            lists[group]['cats'] += [cats.detach().cpu().numpy()]
            lists[group]['gender'] += [data.detach().cpu().numpy()[:, -1]]
            lists[group]['age'] += [data.detach().cpu().numpy()[:, -2]]
            lists[group]['atn'] += [str(x) for x in data.detach().cpu().numpy()[:, -5:-2]]
            lists[group]['inputs'] += [data.view(rec.shape[0], -1).detach().cpu().numpy()]
            lists[group]['encoded_values'] += [enc.detach().cpu().numpy()]
            lists[group]['rec_values'] += [rec.detach().cpu().numpy()]
            try:
                lists[group]['labels'] += [np.array(
                    [self.unique_labels[x] for x in labels.detach().cpu().numpy()])]
            except:
                pass
            traces[group]['acc'] += [np.mean([0 if pred != dom else 1 for pred, dom in
                                              zip(preds.detach().cpu().numpy().argmax(1),
                                                  labels.detach().cpu().numpy())])]
            traces[group]['top3'] += [np.mean([1 if label.item() in pred.tolist()[::-1][:3] else 0 for pred, label in
                                               zip(preds.argsort(1), labels)])]

            traces[group]['closs'] += [classif_loss.item()]
            try:
                traces[group]['mcc'] += [np.round(
                    MCC(labels.detach().cpu().numpy(), preds.detach().cpu().numpy().argmax(1)), 3)
                ]
            except:
                traces[group]['mcc'] = []
                traces[group]['mcc'] += [np.round(
                    MCC(labels.detach().cpu().numpy(), preds.detach().cpu().numpy().argmax(1)), 3)
                ]

            if group in ['train'] and nu != 0:
                # w = np.mean([1/self.class_weights[x] for x in lists[group]['labels'][-1]])
                w = 1
                total_loss = w * nu * classif_loss
                # if self.args.train_after_warmup:
                #     total_loss += rec_loss
                try:
                    total_loss.backward()
                except:
                    pass
                nn.utils.clip_grad_norm_(ae.classifier.parameters(), max_norm=1)
                self.optimizer_c.step()

        return classif_loss, lists, traces

    def forward_discriminate(self, optimizer_b, ae, celoss, loader):
        # Freezing the layers so the batch discriminator can get some knowledge independently
        # from the part where the autoencoder is trained. Only for DANN
        self.freeze_dlayers(ae)
        sampling = True
        for i, batch in enumerate(loader):
            optimizer_b.zero_grad()
            data, meta_inputs, names, labels, domain, to_rec, not_to_rec, pos_batch_sample, \
                neg_batch_sample, meta_pos_batch_sample, meta_neg_batch_sample, set = batch
            # data[torch.isnan(data)] = 0
            data = data.to(self.args.device).float()
            to_rec = to_rec.to(self.args.device).float()
            with torch.no_grad():
                enc, rec, _, kld = ae(data, to_rec, domain, sampling=sampling)
            with torch.enable_grad():
                domain_preds = ae.dann_discriminator(enc)
                bclassif_loss = celoss(domain_preds,
                                       to_categorical(domain.long(), self.n_batches).to(self.args.device).float())
                bclassif_loss.backward()
                nn.utils.clip_grad_norm_(ae.dann_discriminator.parameters(), max_norm=1)
                optimizer_b.step()
        self.unfreeze_layers(ae)
        # return classif_loss, lists, traces

    def get_dloss(self, celoss, domain, domain_preds, set_num=None):
        """
        This function is used to get the domain loss
        Args:
            celoss: PyTorch CrossEntropyLoss instance object
            domain: one-hot encoded domain classes []
            domain_preds: Matrix containing the predicted domains []

        Returns:
            dloss: Domain loss
            domain: True domain (batch) values
        """
        if args.dloss in ['revTriplet', 'revDANN', 'DANN', 'inverseTriplet', 'normae']:
            domain = domain.to(self.args.device).long().to(self.args.device)
            dloss = celoss(domain_preds, domain)
        else:
            dloss = torch.zeros(1)[0].float().to(self.args.device)
        if args.dloss == 'normae':
            dloss = -dloss
        return dloss, domain

    def get_losses(self, scale, smooth, margin, dloss):
        """
        Getter for the losses.
        Args:
            scale: Scaler that was used, e.g. normalizer or binarize
            smooth: Parameter for label_smoothing
            margin: Parameter for the TripletMarginLoss

        Returns:
            sceloss: CrossEntropyLoss (with label smoothing)
            celoss: CrossEntropyLoss object (without label smoothing)
            mseloss: MSELoss object
            triplet_loss: TripletMarginLoss object
        """
        sceloss = nn.CrossEntropyLoss(label_smoothing=smooth)
        celoss = nn.CrossEntropyLoss()
        if self.args.rec_loss == 'mse':
            mseloss = nn.MSELoss()
        elif self.args.rec_loss == 'l1':
            mseloss = nn.L1Loss()
        if scale == "binarize":
            mseloss = nn.BCELoss()
        if dloss == 'revTriplet':
            triplet_loss = nn.TripletMarginLoss(margin, p=2, swap=True)
        else:
            triplet_loss = nn.TripletMarginLoss(0, p=2, swap=False)

        return sceloss, celoss, mseloss, triplet_loss

    def freeze_dlayers(self, ae):
        """
        Freeze all layers except the classifier
        Args:
            ae: AutoEncoder object. It inherits torch.nn.Module

        Returns:
            ae: The same AutoEncoder object, but with all frozen layers. Only the classifier layers are not frozen.

        """
        if not self.args.train_after_warmup:
            for param in ae.dec.parameters():
                param.requires_grad = False
            for param in ae.enc.parameters():
                param.requires_grad = False
            for param in ae.classifier.parameters():
                param.requires_grad = False
            for param in ae.dann_discriminator.parameters():
                param.requires_grad = True
        return ae

    def freeze_ae(self, ae):
        """
        Freeze all layers except the classifier
        Args:
            ae: AutoEncoder object. It inherits torch.nn.Module

        Returns:
            ae: The same AutoEncoder object, but with all frozen layers. Only the classifier layers are not frozen.

        """
        if not self.args.train_after_warmup:
            ae.enc.eval()
            ae.dec.eval()
            for param in ae.dec.parameters():
                param.requires_grad = False
            for param in ae.enc.parameters():
                param.requires_grad = False
            for param in ae.classifier.parameters():
                param.requires_grad = True
            for param in ae.dann_discriminator.parameters():
                param.requires_grad = False
        return ae

    def unfreeze_layers(self, ae):
        """
        Unfreeze all layers
        Args:
            ae: AutoEncoder object. It inherits torch.nn.Module

        Returns:
            ae: The same AutoEncoder object, but with all frozen layers. Only the classifier layers are not frozen.

        """
        for param in ae.parameters():
            param.requires_grad = True
        return ae

    @staticmethod
    def get_mccs(lists, traces):
        """
        Function that gets the Matthews Correlation Coefficients. MCC is a statistical tool for model evaluation.
        It is a balanced measure which can be used even if the classes are of very different sizes.
        Args:
            lists:
            traces:

        Returns:
            traces: Same list as in the inputs arguments, except in now contains the MCC values
        """
        for group in ['train', 'valid', 'test']:
            try:
                preds, classes = np.concatenate(lists[group]['preds']).argmax(1), np.concatenate(
                    lists[group]['classes'])
            except:
                pass
            traces[group]['mcc'] = MCC(preds, classes)

        return traces


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('--random_recs', type=int, default=0)
    parser.add_argument('--predict_tests', type=int, default=0)
    # parser.add_argument('--balanced_rec_loader', type=int, default=0)
    parser.add_argument('--early_stop', type=int, default=50)
    parser.add_argument('--early_warmup_stop', type=int, default=0)
    parser.add_argument('--train_after_warmup', type=int, default=1)
    parser.add_argument('--threshold', type=float, default=0.)
    parser.add_argument('--n_epochs', type=int, default=1000)
    parser.add_argument('--n_trials', type=int, default=100)
    parser.add_argument('--device', type=str, default='cuda:0')
    parser.add_argument('--rec_loss', type=str, default='mse')
    parser.add_argument('--tied_weights', type=int, default=0)
    parser.add_argument('--random', type=int, default=1)
    parser.add_argument('--variational', type=int, default=0)
    parser.add_argument('--zinb', type=int, default=0)  # TODO resolve problems, do not use
    parser.add_argument('--use_valid', type=int, default=0, help='Use if valid data is in a seperate file')  # useless, TODO to remove
    parser.add_argument('--use_test', type=int, default=0, help='Use if test data is in a seperate file')  # useless, TODO to remove
    parser.add_argument('--use_mapping', type=int, default=0, help="Use batch mapping for reconstruct")
    parser.add_argument('--use_gnn', type=int, default=0, help="Use GNN layers")  # useless, TODO to remove
    parser.add_argument('--freeze_ae', type=int, default=0)
    parser.add_argument('--freeze_c', type=int, default=0)
    parser.add_argument('--bdisc', type=int, default=1)
    parser.add_argument('--n_repeats', type=int, default=1)
    parser.add_argument('--dloss', type=str, default='revTriplet')  # one of revDANN, DANN, inverseTriplet, revTriplet
    parser.add_argument('--csv_file', type=str, default='unique_genes.csv')
    parser.add_argument('--best_features_file', type=str, default='')  # best_unique_genes.tsv
    parser.add_argument('--bad_batches', type=str, default='')  # 0;23;22;21;20;19;18;17;16;15
    parser.add_argument('--remove_zeros', type=int, default=1)
    parser.add_argument('--n_meta', type=int, default=0)
    parser.add_argument('--embeddings_meta', type=int, default=0)
    parser.add_argument('--features_to_keep', type=str, default='features_proteins.csv')
    parser.add_argument('--groupkfold', type=int, default=1)
    parser.add_argument('--dataset', type=str, default='prostate')
    parser.add_argument('--bs', type=int, default=32, help='Batch size')
    parser.add_argument('--path', type=str, default='./data/')
    parser.add_argument('--exp_id', type=str, default='default_ae_then_classifier')
    parser.add_argument('--strategy', type=str, default='CU_')
    parser.add_argument('--n_agg', type=int, default=1, help='Number of trailing values to get stable valid values')
    parser.add_argument('--n_layers', type=int, default=2, help='N layers for classifier')
    parser.add_argument('--log1p', type=int, default=1, help='log1p the data? Should be 0 with zinb')

    args = parser.parse_args()

    try:
        mlflow.create_experiment(
            args.exp_id,
            # artifact_location=Path.cwd().joinpath("mlruns").as_uri(),
            # tags={"version": "v1", "priority": "P1"},
        )
    except:
        print(f"\n\nExperiment {args.exp_id} already exists\n\n")
    train = TrainAE(args, args.path, fix_thres=-1, load_tb=False, log_metrics=True, keep_models=False,
                    log_inputs=False, log_plots=True, log_tb=False, log_neptune=False,
                    log_mlflow=True, groupkfold=args.groupkfold, pools=True)

    # train.train()
    # List of hyperparameters getting optimized
    parameters = [
        {"name": "nu", "type": "range", "bounds": [1e-4, 1e2], "log_scale": False},
        {"name": "lr", "type": "range", "bounds": [1e-4, 1e-1], "log_scale": True},
        {"name": "wd", "type": "range", "bounds": [1e-8, 1e-5], "log_scale": True},
        {"name": "l1", "type": "range", "bounds": [1e-8, 1e-5], "log_scale": True},
        # {"name": "lr_b", "type": "range", "bounds": [1e-6, 1e-1], "log_scale": True},
        # {"name": "wd_b", "type": "range", "bounds": [1e-8, 1e-5], "log_scale": True},
        {"name": "smoothing", "type": "range", "bounds": [0., 0.2]},
        {"name": "margin", "type": "range", "bounds": [0., 10.]},
        {"name": "warmup", "type": "range", "bounds": [1, 100]},
        {"name": "disc_b_warmup", "type": "range", "bounds": [1, 2]},

        {"name": "dropout", "type": "range", "bounds": [0.0, 0.5]},
        {"name": "ncols", "type": "range", "bounds": [20, 10000]},
        {"name": "scaler", "type": "choice",
         "values": ['robust', 'binarize', 'standard']},  # scaler whould be no for zinb
        {"name": "layer2", "type": "range", "bounds": [2, 64]},
        {"name": "layer1", "type": "range", "bounds": [2, 256]},
    ]

    # Some hyperparameters are not always required. They are set to a default value in Train.train()
    if args.dloss in ['revTriplet', 'revDANN', 'DANN', 'inverseTriplet', 'normae']:
        # gamma = 0 will ensure DANN is not learned
        parameters += [{"name": "gamma", "type": "range", "bounds": [1e-8, 1e-4], "log_scale": True}]
    if args.variational:
        # beta = 0 because useless outside a variational autoencoder
        parameters += [{"name": "beta", "type": "range", "bounds": [1e-2, 1e2], "log_scale": True}]
    if args.zinb:
        # zeta = 0 because useless outside a zinb autoencoder
        parameters += [{"name": "zeta", "type": "range", "bounds": [1e-2, 1e2], "log_scale": True}]

    best_parameters, values, experiment, model = optimize(
        parameters=parameters,
        evaluation_function=train.train,
        objective_name='mcc',
        minimize=False,
        total_trials=args.n_trials,
        random_seed=41,

    )

    fig = plt.figure()
    # render(plot_contour(model=model, param_x="learning_rate", param_y="weight_decay", metric_name='Loss'))
    # fig.savefig('test.jpg')
    print('Best Loss:', values[0]['loss'])
    print('Best Parameters:')
    print(json.dumps(best_parameters, indent=4))



import torch
import torch.nn as nn
from tqdm import tqdm

from efficient_kan import KAN


def test_mul():
    kan = KAN([2, 2, 1], base_activation=nn.Identity)
    optimizer = torch.optim.LBFGS(kan.parameters(), lr=1)
    with tqdm(range(100)) as pbar:
        for i in pbar:
            loss, reg_loss = None, None

            def closure():
                optimizer.zero_grad()
                x = torch.rand(1024, 2)
                y = kan(x, update_grid=(i % 20 == 0))

                assert y.shape == (1024, 1)
                nonlocal loss, reg_loss
                u = x[:, 0]
                v = x[:, 1]
                loss = nn.functional.mse_loss(y.squeeze(-1), (u + v) / (1 + u * v))
                reg_loss = kan.regularization_loss(1, 0)
                (loss + 1e-5 * reg_loss).backward()
                return loss + reg_loss

            optimizer.step(closure)
            pbar.set_postfix(mse_loss=loss.item(), reg_loss=reg_loss.item())
    for layer in kan.layers:
        print(layer.spline_weight)


"""Training modules for BERNN."""

from .pytorch import KANLinear, KAN
from .train_ae import TrainAE
from .train_ae_classifier_holdout import TrainAEClassifierHoldout
from .train_ae_then_classifier_holdout import TrainAEThenClassifierHoldout

__all__ = [
    "KANLinear",
    "KAN",
    "TrainAE",
    "TrainAEClassifierHoldout",
    "TrainAEThenClassifierHoldout",
]
